﻿namespace MapEditor
{
	using System.Diagnostics;
	using System;
	using CoreLib;
	using CoreLib.Io;

	/// <summary>
	/// Contains the raw data of a block: Its location, tiles, grayscale data, and ownership. This contains no structure whatsoever.
	/// </summary>
	public class ProtoBlock
	{
		#region Constuctors
		/// <summary>
		/// Default .ctor.
		/// </summary>
		public ProtoBlock()
		{
			this.TileData = new Tile();
			this.Grayscale = new ColoringData(true);
			this.Ownership = new OwnershipData(true);
		}
		#endregion

		#region Data storage
		/// <summary>
		/// Gets or sets the x-coordinate of the block's offset.
		/// </summary>
		public short XOffset
		{
			get;
			set;
		}

		/// <summary>
		/// Gets or sets the y-coordinate of the block's offset.
		/// </summary>
		public short YOffset
		{
			get;
			set;
		}

		/// <summary>
		/// Gets the tile data of the block.
		/// </summary>
		public Tile TileData
		{
			get;
			private set;
		}

		/// <summary>
		/// Gets the grayscale data of the block.
		/// </summary>
		public ColoringData Grayscale
		{
			get;
			private set;
		}

		/// <summary>
		/// Gets the ownership data of the block.
		/// </summary>
		public OwnershipData Ownership
		{
			get;
			private set;
		}
		#endregion

		/* Data Format
		 * int length; //Length of block, does not include this
		 * short xOffset; //In blocks
		 * short yOffset; //In blocks
		 * Tile tile; //Variable, use tile's reader/writer
		 * GrayscaleData grayscale; //Variable, use grayscale's reader/writer
		 * BlockOwnershipData ownership; //Variable, use ownership's reader/writer
		 */

		#region Streaming
		public static ProtoBlock Read(IStreamReader dataStream, IStreamReader headerStream)
		{
			ProtoBlock ret = new ProtoBlock();

			ret.XOffset = headerStream.ReadInt16();
			ret.YOffset = headerStream.ReadInt16();
			int position = headerStream.ReadInt32();
			int length = headerStream.ReadInt32();

			dataStream.Position = position;

			ret.TileData = Tile.Read(dataStream);
			ret.Grayscale = ColoringData.Read(dataStream);
			ret.Ownership = OwnershipData.Read(dataStream);

			return ret;
		}
		public static ProtoBlock Read(IStreamReader dataStream, BlockHolder header)
		{
			dataStream.Position = header.Position;

			ProtoBlock ret = new ProtoBlock();
			ret.XOffset = (short)header.XOffset;
			ret.YOffset = (short)header.YOffset;

			ret.TileData = Tile.Read(dataStream);
			ret.Grayscale = ColoringData.Read(dataStream);
			ret.Ownership = OwnershipData.Read(dataStream);

			return ret;
		}

		/// <summary>
		/// Reads the stream when the length field is included.
		/// </summary>
		/// <param name="stream">The stream to read from.</param>
		/// <returns>A ProtoBlock containing the read data.</returns>
		[Obsolete]
		public static ProtoBlock ReadWithLengthField(ByteReader stream)
		{
			stream.Position += 4;
			return ReadWithoutLengthField(stream);
		}

		/// <summary>
		/// Reads the stream after the length field has already passed.
		/// </summary>
		/// <param name="stream">The stream to read from.</param>
		/// <returns>A ProtoBlock containing the read data.</returns>
		[Obsolete]
		public static ProtoBlock ReadWithoutLengthField(ByteReader stream)
		{
			ProtoBlock ret = new ProtoBlock();
			ret.XOffset = stream.ReadInt16();
			ret.YOffset = stream.ReadInt16();
			ret.TileData = Tile.ReadStream(stream);
			ret.Grayscale = ColoringData.Read(stream);
			ret.Ownership = OwnershipData.Read(stream);
			return ret;
		}

		public void Write(IStreamWriter dataStream, IStreamWriter headerStream)
		{
			int begin = dataStream.Position;
			this.TileData.WriteStream(dataStream);
			this.Grayscale.Write(dataStream);
			this.Ownership.Write(dataStream);
			int end = dataStream.Position;

			headerStream.WriteInt16(this.XOffset);
			headerStream.WriteInt16(this.YOffset);
			headerStream.WriteInt32(begin);
			headerStream.WriteInt32(end - begin);
		}

		/// <summary>
		/// Writes data to a stream.
		/// </summary>
		/// <param name="stream">The stream to write to.</param>
		[Obsolete]
		public void Write(IStreamWriter stream)
		{
			int pos = stream.Position;
			int length = 4;
			stream.WriteInt32(0);
			stream.WriteInt16(this.XOffset); // Length 2
			stream.WriteInt16(this.YOffset); // Length 2
			length += this.TileData.WriteStream(stream); // 1 + HasTile?1024*4:0
			length += this.Grayscale.Write(stream); // 1 + Compressed?1:1024
			length += this.Ownership.Write(stream); // Whatever it gives
			int nowPos = stream.Position;
			stream.Position = pos;
			stream.WriteInt32(length);
			stream.Position = nowPos;
		}
		#endregion
	}
}