﻿using System;
using System.Collections.Generic;
using System.IO;
using CoreLib;
using CoreLib.Io;
using Leaf = MapUtility.Leaf;
using Rect = System.Drawing.Rectangle;

using MapUtility;

namespace MapEditor
{
	internal static class BoundBoxMaker
	{
		private static readonly int MapWidth = Settings.Lm1Width;
		private static readonly int MapHeight = Settings.Lm1Height;
		private static readonly int BlockSize = Global.BlockHeight;

		/// <summary>
		/// Creates the boundbox.tbl file
		/// </summary>
		internal static void MakeBoundBoxes()
		{
			int maxProvNumber = Settings.QueryMaxProvinceNumber();
			Console.Write("Loading boxes...");
			List<Box> provinces = new List<Box>(maxProvNumber);
			Box defaultBox = new Box(Width, Height, 0, 0);
			Iteration.Repeat(provinces, list => list.Add(defaultBox), maxProvNumber);
			Console.WriteLine("Done!");
			Console.Write("Loading lightmap...");
			Lightmap lm = new Lightmap(LightmapOption.LightMap1, Shared.OutputFolder);
			Console.WriteLine("Done!");

			Console.Write("Doing row ");
			for (int y = 0; y < MapHeight; y++)
			{
				string val = y.ToString();
				Console.Write(val);
				for (int x = 0; x < MapWidth; x++)
				{
					MapBlock curBlock = lm[x, y];
					curBlock.EnsureDecompressed();
					BoundsFromLeaf(provinces, curBlock);
					curBlock.DiscardData();
				}
				ConsoleEx.Backspace(val.Length);
			}
			Console.WriteLine("...Done!");

			#region old
			/*
			using (ByteReader reader = new ByteReader(Shared.Lightmap1RadishPath))
			{
				Console.Write("Doing row ");
				for (int blockY = 0; blockY < MapHeight; blockY++)
				{
					string val = blockY.ToString();
					Console.Write(val);
					for (int blockX = 0; blockX < MapWidth; blockX++)
					{
						OwnershipData ownership =
							RadishHelper.LoadBlockData(blockX, blockY, Shared.Lightmap1RadishHeaders, reader).Ownership;
						if (ownership.Compressed)
						{
							BoundFromCompressed(provinces, blockY, blockX, ownership[0].OwnerId);
						}
						else
						{
							BoundsFromData(provinces, blockY, blockX, ownership);
						}
					}
					ConsoleEx.Backspace(val.Length);
				}
				Console.WriteLine("...Done!");
			}
			*/
			#endregion

			WriteBoundBoxes(maxProvNumber, provinces);
		}
		static void BoundsFromLeaf(List<Box> provinces, Leaf leaf)
		{
			if (leaf.HasChildren)
			{
				BoundsFromLeaf(provinces, leaf.Children[0]);
				BoundsFromLeaf(provinces, leaf.Children[1]);
				BoundsFromLeaf(provinces, leaf.Children[2]);
				BoundsFromLeaf(provinces, leaf.Children[3]);
			}
			else
			{
				Rect leafBounds = new System.Drawing.Rectangle(leaf.xCoordinate, leaf.yCoordinate, leaf.Shift, leaf.Shift);

				Box box = provinces[leaf.OwnerId.OwnerId];

				if (box.UpperLeftX > leafBounds.X)
					box.UpperLeftX = (short)leafBounds.X;
				if (box.UpperLeftY > leafBounds.Y)
					box.UpperLeftY = (short)leafBounds.Y;
				if (box.LowerRightX < leafBounds.Right)
					box.LowerRightX = (short)leafBounds.Right;
				if (box.LowerRightY < leafBounds.Bottom)
					box.LowerRightY = (short)leafBounds.Bottom;

				provinces[leaf.OwnerId.OwnerId] = box;
			}
		}

		private static void BoundsFromData(List<Box> provinces, int blockY, int blockX, OwnershipData ownership)
		{
			for (int x = 0; x < BlockSize; x++)
			{
				for (int y = 0; y < BlockSize; y++)
				{
					int shiftX = blockX * BlockSize + x;
					int shiftY = blockY * BlockSize + y;
					int id = ownership[x, y].OwnerId;
					if (id == 0)
						continue;
					Box box = provinces[id];
					if (box.UpperLeftX > shiftX)
						box.UpperLeftX = (short)shiftX;
					if (box.UpperLeftY > shiftY)
						box.UpperLeftY = (short)shiftY;
					if (box.LowerRightX < shiftX)
						box.LowerRightX = (short)shiftX;
					if (box.LowerRightY < shiftY)
						box.LowerRightY = (short)shiftY;
					provinces[id] = box;
				}
			}
		}
		private static void WriteBoundBoxes(int maxProvNumber, List<Box> provinces)
		{
			Console.WriteLine("Creating boundbox.tbl...");
			using (BinaryStream bw = new BinaryStream(Shared.BoundboxPath, FileMode.Create))
			{
				//Provinces here
				for (int i = 0; i < maxProvNumber; i++)
				{
					bw.WriteInt32(provinces[i].UpperLeftX);
					bw.WriteInt32(provinces[i].UpperLeftY);
					bw.WriteInt32(provinces[i].LowerRightX);
					bw.WriteInt32(provinces[i].LowerRightY);
				}
			}
			Console.WriteLine("Done!");
		}
		private static void BoundFromCompressed(List<Box> provinces, int blockY, int blockX, int id)
		{
			int upperX = blockX * BlockSize;
			int upperY = blockY * BlockSize;
			int lowerX = upperX + BlockSize;
			int lowerY = upperY + BlockSize;
			Box box = provinces[id];
			if (box.UpperLeftX > upperX)
				box.UpperLeftX = (short)upperX;
			if (box.UpperLeftY > upperY)
				box.UpperLeftY = (short)upperY;
			if (box.LowerRightX < lowerX)
				box.LowerRightX = (short)lowerX;
			if (box.LowerRightY < lowerY)
				box.LowerRightY = (short)lowerY;
			provinces[id] = box;
		}

		private const int Zero = 0;
		private static readonly short Width = (short)(Settings.Lm1Width * Global.BlockWidth);
		private static readonly short Height = (short)(Settings.Lm1Height * Global.BlockHeight);
	}
}