﻿using System;
using System.Collections.Generic;
using CoreLib;
using CoreLib.Io;
using MapUtility;
using Leaf = MapUtility.Leaf;

namespace MapEditor
{
	public static class LightmapMaker
	{
		private static ColoringData grayscale;
		private static OwnershipData ownershipBlock;
		private static void LoadData()
		{
			Console.Write("Loading data...");
			try
			{
				Global.AdjacentTable.GetHashCode(); //To force it to report an error NOW
			}
			catch (Exception e)
			{
				if (e.Message != "Missing file!")
					throw;
			}
			Console.WriteLine("Done!");
		}
		private static List<PixelOwnership> GetProvinceList()
		{
			List<PixelOwnership> ret = new List<PixelOwnership>();

			if (ownershipBlock.Compressed)
			{
				ret.Add(ownershipBlock[0]);
				return ret;
			}

			for (int i = 0; i < 1024; i++)
			{
				if (!ret.Contains(ownershipBlock[i]))
					ret.Add(ownershipBlock[i]);
			}
			PixelOwnership temp;
			for (int i = 0; i < ret.Count; i++)
			{
				if (ret[i].IsBorder)
				{
					temp = ret[i];
					temp.IsBorder = false;
					if (ret.IndexOf(temp) == -1) //The ID was not found!!!
					{
						temp = new PixelOwnership();
						temp.OwnerId = ret[i].OwnerId;
						ret.Add(temp); //A dummy ID
					}
				}
			}
			ret.Sort();
			return ret;
		}

		private static void MakeLightMap(string path, int xblocks, int yblocks, IStreamReader readstream, BlockHolder[,] blocks)
		{
			Console.WriteLine("Making " + path);

			Lightmap final = new Lightmap(xblocks, yblocks, path);

			Console.Write("Doing row: ");
			for (int y = 0; y < yblocks; y++)
			{
				string row = y.ToString();
				Console.Write(row);
				for (int x = 0; x < xblocks; x++)
				{
					ProtoBlock current = RadishHelper.LoadBlockData(x, y, blocks, readstream);

					grayscale = current.Grayscale;
					ownershipBlock = current.Ownership;

					List<PixelOwnership> provinces = GetProvinceList(); //This will be written to the file

					byte[,] grayscaleVals;

					int rightX = x < xblocks - 1 ? x + 1 : 0;
					ProtoBlock right = RadishHelper.LoadBlockData(rightX, y, blocks, readstream);
					int? bottomY = y < yblocks - 1 ? y + 1 : (int?)null;
					if (bottomY == null)
					{
						grayscaleVals = new byte[33, 32];
					}
					else
					{
						grayscaleVals = new byte[33, 33];
						ProtoBlock bottom = RadishHelper.LoadBlockData(x, bottomY.Value, blocks, readstream);
						for (int i = 0; i < 32; i++) //Fill info from bottom block
							grayscaleVals[i, 32] = bottom.Grayscale[i, 0];
						ProtoBlock bottomRight = RadishHelper.LoadBlockData(rightX, bottomY.Value, blocks, readstream);
						grayscaleVals[32, 32] = bottomRight.Grayscale[0, 0]; //Fill bottomright block
					}
					for (int i = 0; i < 32; i++) //Fill info from block on right
					{
						grayscaleVals[32, i] = right.Grayscale[0, i];
					}

					LoadMainGrayscale(current, grayscaleVals);

					MapBlock block = new MapBlock((ushort)x, (ushort)y, final);
					final[x, y] = block;

					//Load province ID list
					block.SetProvinceIds(provinces);

					CompressBlock(block, grayscaleVals, ownershipBlock, 0, 0);

					final[x, y].FinishedToRaw();
					final[x, y].DiscardData();
				}
				ConsoleEx.Backspace(row.Length);
			}
			Console.WriteLine();
			Console.Write("Saving...");

			final.Write();
			Console.WriteLine("Done!");

			Console.WriteLine("Done making " + path);
		}

		private static void LoadMainGrayscale(ProtoBlock current, byte[,] grayscaleVals)
		{
			if (current.Grayscale.Compressed)
			{
				byte value = current.Grayscale[0];
				for (int xx = 0; xx < 32; xx++)
				{
					for (int yy = 0; yy < 32; yy++)
					{
						grayscaleVals[xx, yy] = value;
					}
				}
			}
			else
			{
				for (int xx = 0; xx < 32; xx++)
				{
					for (int yy = 0; yy < 32; yy++)
					{
						grayscaleVals[xx, yy] = current.Grayscale[xx, yy];
					}
				}
			}
		}
		private static void CompressBlock(Leaf current, byte[,] availableGrayscale, OwnershipData ownership, int xStart, int yStart)
		{
			if (ownership.AllIsSame(xStart, yStart, current.Shift) &&
				ShadingIsCorrect(current.Shift, xStart, yStart, availableGrayscale))
			{
				current.OwnerId = ownership[xStart, yStart];
				current.GrayscaleColor = availableGrayscale[xStart, yStart];
			}
			else
			{
				for (int i = 0; i < 4; i++)
				{
					Leaf l = new Leaf(current, i);
					current.AddChild(l);
					CompressBlock(l, availableGrayscale, ownership,
						i == 0 || i == 2 ? xStart + l.Shift : xStart,
						i == 0 || i == 1 ? yStart + l.Shift : yStart);
				}
			}
		}
		private static bool ShadingIsCorrect(int shift, int xStart, int yStart, byte[,] availableGrayscale)
		{
			if (shift == 1)
				return true;
			if (yStart + shift == availableGrayscale.GetLength(1))
			{
				//Perhaps we should think about ocean on the bottom?
				return false;
			}

			byte[,] data = new byte[2, shift + 1];
			data[0, 0] = availableGrayscale[xStart, yStart];
			data[0, shift] = availableGrayscale[xStart, yStart + shift];
			data[1, shift] = availableGrayscale[xStart + shift, yStart + shift];
			data[1, 0] = availableGrayscale[xStart + shift, yStart];

			//int realMaxDiff = MaxDiff;
			//Is int because shifts can be negative, and sbytes might not have enough range
			int firstShift = data[0, shift] - data[0, 0];
			int secondShift = data[1, shift] - data[1, 0];
			for (int i2 = 1; i2 < shift; i2++)
			{
				data[0, i2] = (byte)(data[0, 0] + (firstShift * i2 + 0.5) / shift);
				data[1, i2] = (byte)(data[1, 0] + (secondShift * i2 + 0.5) / shift);
			}

			for (int row = 0; row < shift; row++)
			{
				int valShift = data[1, row] - data[0, row];
				for (int i2 = 0; i2 < shift; i2++)
				{
					byte value = (byte)(data[0, row] + (valShift * i2 + 0.5) / shift);
					byte wanted = availableGrayscale[xStart + i2, yStart + row];
					if (!(value >= wanted - 1 && value <= wanted + 1))
						return false;
				}
			}

			return true;
		}

		/// <summary>
		/// This searches the temp folder for radish files
		/// </summary>
		public static void MakeLightmaps()
		{
			LoadData();
			using (BinaryStream stream = new BinaryStream(Shared.Lightmap1RadishPath, System.IO.FileMode.Open))
			{
				MakeLightMap(Shared.Lightmap1Path, Settings.Lm1Width, Settings.Lm1Height, stream, Shared.Lightmap1RadishHeaders);
			}
			using (BinaryStream stream = new BinaryStream(Shared.Lightmap2RadishPath, System.IO.FileMode.Open))
			{
				MakeLightMap(Shared.Lightmap2Path, Settings.Lm2Width, Settings.Lm2Height, stream, Shared.Lightmap2RadishHeaders);
			}
			using (ByteReader stream = new ByteReader(Shared.Lightmap3RadishPath, true))
			{
				MakeLightMap(Shared.Lightmap3Path, Settings.Lm3Width, Settings.Lm3Height, stream, Shared.Lightmap3RadishHeaders);
			}
			if (Settings.HasFourLightmaps)
			{
				using (ByteReader stream = new ByteReader(Shared.Lightmap4RadishPath, true))
				{
					MakeLightMap(Shared.Lightmap4Path, Settings.Lm4Width, Settings.Lm4Height, stream, Shared.Lightmap4RadishHeaders);
				}
			}
			Clean();
		}

		private static void Clean()
		{
			grayscale = null;
			ownershipBlock = null;
		}

		#region Junk
		/*private static void CompressBlock(ref int indexPos, int[] ownershipFinal, byte[] grayscaleFinal, List<bool> layout,
			byte[] ownership, ColoringData strengths, int zoom, int x, int y)
		{
			int shift = 1 << zoom;
			if (AllIsSame(ownership, x, y, shift) && strengths.AllIsSame(x, y, shift))
			{
				layout.Add(false);
				ownershipFinal[indexPos] = ownership[y * 32 + x];
				grayscaleFinal[indexPos] = grayscale[y * 32 + x];
				indexPos++;
			}
			else
			{
				layout.Add(true);
				if (zoom > 1)
				{
					shift >>= 1;
					//BR -> BL -> TR -> TL
					CompressBlock(ref indexPos, ownershipFinal, grayscaleFinal, layout, ownership, strengths, zoom - 1, x + shift, y + shift);
					CompressBlock(ref indexPos, ownershipFinal, grayscaleFinal, layout, ownership, strengths, zoom - 1, x, y + shift);
					CompressBlock(ref indexPos, ownershipFinal, grayscaleFinal, layout, ownership, strengths, zoom - 1, x + shift, y);
					CompressBlock(ref indexPos, ownershipFinal, grayscaleFinal, layout, ownership, strengths, zoom - 1, x, y);
				}
				else
				{
					ownershipFinal[indexPos] = ownership[(y + 1) * 32 + x + 1];
					grayscaleFinal[indexPos] = grayscale[(y + 1) * 32 + x + 1];
					indexPos++;
					ownershipFinal[indexPos] = ownership[(y + 1) * 32 + x];
					grayscaleFinal[indexPos] = grayscale[(y + 1) * 32 + x];
					indexPos++;
					ownershipFinal[indexPos] = ownership[y * 32 + x + 1];
					grayscaleFinal[indexPos] = grayscale[y * 32 + x + 1];
					indexPos++;
					ownershipFinal[indexPos] = ownership[y * 32 + x];
					grayscaleFinal[indexPos] = grayscale[y * 32 + x];
					indexPos++;
				}
			}
		}*/
		/*private static void WrshipData(BinaryStream stream, int indexPos, int[] ownershipFinal, int ownershipIndexWidth, int x)
		{
			if (ownershipIndexWidth == 0) return;
			int AndVal = (1 << ownershipIndexWidth) - 1;
			int shiftPos = 0;
			bool needWrite = false;
			byte data = 0;
			for (int i = 0; i < indexPos; i++)
			{
				data |= (byte)((ownershipFinal[i] & AndVal) << shiftPos);
				shiftPos += ownershipIndexWidth;
				if (shiftPos == 8)
				{
					stream.WriteByte(data);
					data = 0;
					shiftPos = 0;
					needWrite = false;
				}
				else
					needWrite = true;
			}
			if (needWrite)
			{
				stream.WriteByte(data);
			}
		}*/
		/*private static void WriteGrayscaleData(BinaryStream stream, int indexPos, byte[] grayscaleFinal)
		{
			byte first = 0;
			byte second = 0;
			byte third = 0;
			bool written = false;
			for (int i = 0; i < indexPos; i++)
			{
				switch (i % 4)
				{
					case 0:
						first = grayscaleFinal[i];
						written = false;
						break;
					case 1:
						first |= (byte)((grayscaleFinal[i] & 3) << 6);
						second |= ((byte)(grayscaleFinal[i] >> 2));
						break;
					case 2:
						second |= (byte)((grayscaleFinal[i] & 15) << 4);
						third |= ((byte)(grayscaleFinal[i] >> 4));
						break;
					case 3:
						third |= (byte)(grayscaleFinal[i] << 2);
						stream.WriteBytes(first, second, third);
						written = true;
						first = second = third = 0;
						break;
				}
			}
			if (!written)
				stream.WriteBytes(first, second, third);
		}*/
		/*private static byte[] LoadOwnership(List<PixelOwnership> provinces)
		{
			byte[] ownership = new byte[1024];
			for (int iteration = 0; iteration < 1024; iteration++)
			{
				ownership[iteration] = (byte)provinces.IndexOf(ownershipBlock[iteration]);
			}
			return ownership;
		}*/
		/*private static int GetOwnershipWidth(int provinceCount)
		{
			if (provinceCount == 1)
				return 0;
			if (provinceCount == 2)
				return 1;
			if (provinceCount <= 4)
				return 2;
			if (provinceCount <= 16)
				return 4;
			if (provinceCount <= 256)
				return 8;
			throw new Exception();
		}*/
		/*private static bool AllIsSame<T>(T[] array, int x, int y, int shift)
		{
			T toCompare = array[y * 32 + x];
			for (int xPos = x; xPos < x + shift; xPos++)
				for (int yPos = y; yPos < y + shift; yPos++)
				{
					if (!array[yPos * 32 + xPos].Equals(toCompare))
						return false;
				}
			return true;
		}*/
		#endregion
	}
}