﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Drawing;
using System.Linq;
using CoreLib;
using CoreLib.Graphics;
using CoreLib.Graphics.Photoshop;
using CoreLib.Io;

namespace MapEditor
{
	public abstract class LineData
	{
		/// <summary>
		/// The extension of the file.
		/// </summary>
		public string Extension { get; private set; }
		/// <summary>
		/// The original line.
		/// </summary>
		public string Line { get; private set; }
		/// <summary>
		/// The original arguments that were passed in.
		/// </summary>
		public ReadOnlyCollection<string> Arguments { get; private set; }
		public int xOffset { get; protected set; } //The picture's topleft xOffset in blocks
		public int yOffset { get; protected set; } //The picture's topleft yOffset in blocks
		public bool HasTilesLayer { get; protected set; }
		public bool HasSpecialBaseLayer { get; protected set; }
		public abstract FileTypes FileType { get; }
		/// <summary>
		/// A list of points representing the blocks to be recorded.
		/// The offset is in blocks from the top-left of the image.
		/// This will be null if every block is used.
		/// </summary>
		public ReadOnlyCollection<Point> BaseLayerPoints { get; protected set; }

		#region File paths
		public string FilePath { get; private set; }
		public abstract string GrayscalePath { get; }
		public abstract string ProvinceOwnershipPath { get; }
		public abstract string TilesPath { get; }
		public abstract string BordersPath { get; }
		#endregion

		protected LineData(string line)
		{
			this.Line = line;
			this.Arguments = new ReadOnlyCollection<string>(line.Split(';'));
			FilePath = Arguments[0];
			Extension = FilePath.Split('.').Last();
		}
		public static LineData Interpret(string line)
		{
			string[] args = line.Split(';');
			string extension = args[0].Split('.').Last();
			switch (extension.ChangeToLower())
			{
				case "psd":
					return new PhotoshopData(line);
				case "bmp":
				case "gif":
				case "jpg":
				case "png":
				case "tiff":
				case "wmf":
				case "exif":
				case "emf":
					return new SimpleImageData(line);
				default:
					throw new NotSupportedException("Extension not supported: " + extension);
			}
		}

		protected List<Point> GetRecordedTiles(ImageData<GrayscaleData> baseImage)
		{
			List<Point> recordedTiles = new List<Point>();

			for (int x = 0; x < baseImage.Width; x += 32) //Use basePic to get what tiles should be copied
				for (int y = 0; y < baseImage.Height; y += 32) //Question: Do we need to compensate for the shift from the main pic?
					if (baseImage[x, y].Strength == 0)
						recordedTiles.Add(new Point(x / 32, y / 32));

			return recordedTiles;
		}
	}

	public sealed class PhotoshopData : LineData
	{
		public override string BordersPath
		{
			get
			{
				return FilePath;
			}
		}
		public override string GrayscalePath
		{
			get
			{
				return FilePath;
			}
		}
		public override string ProvinceOwnershipPath
		{
			get
			{
				return FilePath;
			}
		}
		public override string TilesPath
		{
			get
			{
				return FilePath;
			}
		}

		public const FileTypes Type = FileTypes.Photoshop;
		public override FileTypes FileType
		{
			get { return Type; }
		}

		public PhotoshopData(string line)
			: base(line)
		{
			FileReader psd = new FileReader(FilePath);

			//Load offsets
			if (Arguments.Count >= 3) //Offsets listed here
			{
				xOffset = int.Parse(Arguments[1]);
				yOffset = int.Parse(Arguments[2]);
			}
			else
			{
				string[] offsetData = psd.Resources.GetResource<FileInfoResource>().GetInfo(FileInfoResource.City).Split(';');
				xOffset = int.Parse(offsetData[0]);
				yOffset = int.Parse(offsetData[1]);
			}

			HasTilesLayer = psd.Layers[Shared.TilesImageName] != null;

			//Load recorded points
			Layer baseLayer = psd.Layers[Shared.BaseImageName] as Layer;
			if (baseLayer == null)
			{
				HasSpecialBaseLayer = false;
			}
			else
			{
				HasSpecialBaseLayer = true;
				baseLayer.LoadPic(ColorType.Grayscale);
				BaseLayerPoints = new ReadOnlyCollection<Point>(GetRecordedTiles(baseLayer.PictureData.CastAs<GrayscaleData>()));
			}
		}
	}
	public sealed class SimpleImageData : LineData
	{
		private string bordersPath;
		public override string BordersPath
		{
			get
			{
				return bordersPath;
			}
		}
		private string grayscalePath;
		public override string GrayscalePath
		{
			get
			{
				return grayscalePath;
			}
		}
		private string provinceOwnershipPath;
		public override string ProvinceOwnershipPath
		{
			get
			{
				return provinceOwnershipPath;
			}
		}
		private string tilesPath;
		public override string TilesPath
		{
			get
			{
				return tilesPath;
			}
		}

		public const FileTypes Type = FileTypes.SimpleImages;
		public override FileTypes FileType
		{
			get { return Type; }
		}

		public SimpleImageData(string line)
			: base(line)
		{
			//Load offsets
			xOffset = int.Parse(Arguments[1]);
			yOffset = int.Parse(Arguments[2]);

			string[] separatedbydot = FilePath.Split('.');
			string firstPart = string.Join(".", separatedbydot, 0, separatedbydot.Length - 1) + ".";
			string lastPart = "." + Extension;
			bordersPath = firstPart + Shared.BordersImageName + lastPart;
			grayscalePath = firstPart + Shared.GrayscaleImageName + lastPart;
			provinceOwnershipPath = firstPart + Shared.ProvIdImageName + lastPart;
			tilesPath = firstPart + Shared.TilesImageName + lastPart;

			//Has tiles layer?
			HasTilesLayer = File.Exists(tilesPath);

			//Load recorded points
			string basePath = firstPart + Shared.BaseImageName + lastPart;
			if (File.Exists(basePath))
			{
				HasSpecialBaseLayer = true;
				BaseLayerPoints = new ReadOnlyCollection<Point>(GetRecordedTiles(ImageData<GrayscaleData>.ToData(new Bitmap(basePath))));
			}
			else
				HasSpecialBaseLayer = false;
		}
	}

	public enum FileTypes
	{
		Photoshop,
		SimpleImages,
		Pcx,
		Raw,
		Svg,
		Pdn
	}
}