﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using CoreLib.Io;
using MapUtility;

namespace MapEditor
{
	public sealed class OwnershipData : CompressableBlockBase<PixelOwnership>
	{
		#region Constructors
		public OwnershipData(bool compressed) : base(compressed) { }
		//Compressed, gives value
		public OwnershipData(PixelOwnership value) : base(value) { }
		public OwnershipData(PixelOwnership[] values) : base(values) { }
		public OwnershipData(PixelOwnership[] values, bool checkCompression) : base(values, checkCompression) { }
		#endregion

		#region Streaming
		public unsafe static OwnershipData Read(ByteReader stream)
		{
			if (stream.ReadBool())
			{
				return new OwnershipData(PixelOwnership.Read(stream));
			}
			else
			{
				PixelOwnership[] vals = new PixelOwnership[1024];
				byte[] source = stream.ReadBytes(PixelOwnership.Length * 1024);
				fixed (byte* p = &source[0])
				{
					PixelOwnership* ptr = (PixelOwnership*)p;
					for (int i = 0; i < 1024; i++)
					{
						vals[i] = ptr[i];
					}
				}
				return new OwnershipData(vals);
			}
		}
		public unsafe static OwnershipData Read(IStreamReader stream)
		{
			if (stream.ReadBool())
			{
				return new OwnershipData(PixelOwnership.Read(stream));
			}
			else
			{
				PixelOwnership[] vals = new PixelOwnership[1024];
				byte[] source = stream.ReadBytes(PixelOwnership.Length * 1024);
				fixed (byte* sourcePtr = &source[0])
				{
					fixed (PixelOwnership* destPtr = &vals[0])
					{
						CoreLib.Interop.Win32.Kernel.CopyMemory((IntPtr)destPtr, (IntPtr)sourcePtr, 1024 * PixelOwnership.Length);
					}
					/*PixelOwnership* ptr = (PixelOwnership*)p;
					for (int i = 0; i < 1024; i++)
					{
						vals[i] = ptr[i];
					}*/
				}
				return new OwnershipData(vals);
			}
		}
		public unsafe int Write(IStreamWriter stream) //Returns length
		{
			TryToCompress();
			stream.WriteBool(Compressed);
			if (Compressed)
			{
				stream.Write(value);
				return 6; //1 is the flag
			}
			else
			{
				byte[] buffer = new byte[PixelOwnership.Length * data.Length];
				fixed (PixelOwnership* srcPtr = &data[0])
				{
					fixed (byte* destPtr = &buffer[0])
					{
						CoreLib.Interop.Win32.Kernel.CopyMemory((IntPtr)destPtr, (IntPtr)srcPtr, buffer.Length);
					}
				}
				stream.WriteBytes(buffer);
				/*foreach (PixelOwnership pos in data)
				{
					stream.Write(pos);
				}*/
				return 1 + 5 * data.Length;
			}
		}
		#endregion

		#region Misc Methods
		public void RemoveBorders()
		{
			if (Compressed)
				value.IsBorder = false;
			else
			{
				for (int i = 0; i < 1024; i++)
				{
					data[i].IsBorder = false;
				}
			}
		}
		#endregion

		#region Removed Members
		/* Has no callers
		[DebuggerHidden]
		public RgbData[,] ToColorData()
		{
			RgbData[,] ret = new RgbData[32, 32];
			for (int i = 0; i < 1024; i++)
			{
				ret[i / 32, i % 32] = data[i].ToColor();
			}
			return ret;
		}*/
		#endregion
	}
	public abstract class CompressableBlockBase<T> : IEnumerable<T> where T : IEquatable<T>
	{
		#region Constructors
		protected CompressableBlockBase(bool compressed)
		{
			if (!compressed)
				data = new T[1024];
		}
		protected CompressableBlockBase(T value)
		{
			this.value = value;
		}
		protected CompressableBlockBase(T[] values, bool checkCompression)
		{
			LoadDataFrom(values, checkCompression);
		}
		protected CompressableBlockBase(T[] values) : this(values, false) { }
		#endregion

		#region From
		public void LoadDataFrom(T[] values, bool checkCompression)
		{
			if (values.Length != 1024)
				throw new Exception("values must be 1024 long!");
			data = values;
			if (checkCompression)
				TryToCompress();
		}
		#endregion

		#region Data
		protected T[] data;
		protected T value;
		public bool Compressed
		{
			[DebuggerHidden]
			get { return data == null; }
		}
		#endregion

		#region Misc Methods
		/// <summary>
		/// Attempts to compresss all the data into a single member. This only works if all tiles of the data are the same.
		/// </summary>
		public void TryToCompress()
		{
			if (Compressed) return;
			T val = data[0];
			bool allAreEqual = true;
			for (int i = 0; i < 1024; i++)
			{
				if (!data[i].Equals(val))
				{
					allAreEqual = false;
					break;
				}
			}
			if (allAreEqual)
			{
				data = null;
				value = val;
			}
		}
		/// <summary>
		/// Checks whether all tiles are the same in the specified region.
		/// </summary>
		/// <param name="x">The x-coordinate of the top-left of the region.</param>
		/// <param name="y">The y-coordinate of the top-left of the region.</param>
		/// <param name="shift">The height and width of the region.</param>
		/// <returns>Returns true if all the items in the specified regions are the same, otherwise false.</returns>
		public bool AllIsSame(int x, int y, int shift)
		{
			if (Compressed)
				return true;

			T toCompare = data[y * 32 + x];
			for (int xPos = x; xPos < x + shift; xPos++)
				for (int yPos = y; yPos < y + shift; yPos++)
				{
					if (!data[yPos * 32 + xPos].Equals(toCompare))
						return false;
				}
			return true;
		}
		/// <summary>
		/// Sets all values to the specified value.
		/// </summary>
		/// <param name="value">The value to set everything to.</param>
		public void SetAll(T value)
		{
			this.value = value;
			this.data = null;
		}
		#endregion

		#region Data accessors
		/// <summary>
		/// Returns the T at the specified index.
		/// </summary>
		/// <param name="index">The index where the requested T is.</param>
		/// <returns>The T that is searched for.</returns>
		public T this[int index]
		{
			[DebuggerHidden]
			get
			{
				if (Compressed)
					return value;
				return data[index];
			}
			set
			{
				if (Compressed)
				{
					if (!value.Equals(this.value))
					{
						data = new T[1024];
						for (int i = 0; i < 1024; i++)
						{
							data[i] = this.value;
						}
					}
					else
						return;
				}
				data[index] = value;
			}
		}
		/// <summary>
		/// Returns the T item at the requested X and Y coordinates.
		/// </summary>
		/// <param name="x">The x-coordinate of the requested item.</param>
		/// <param name="y">The y-coordinate of the requested item.</param>
		/// <returns>The requested item.</returns>
		public T this[int x, int y]
		{
			[DebuggerHidden]
			get
			{
				return this[y * 32 + x];
			}
			set
			{
				this[y * 32 + x] = value;
			}
		}
		#endregion

		#region IEnumerable<T> Members
		IEnumerator IEnumerable.GetEnumerator()
		{
			if (Compressed)
			{
				return GetSingleEnumerator();
				//yield break;
				//T[] arr = new T[1];
				//arr[0] = value;
				//return arr.GetEnumerator();
			}
			else
			{
				return data.GetEnumerator();
			}
		}
		private IEnumerator GetSingleEnumerator()
		{
			yield return value;
		}
		public IEnumerator<T> GetEnumerator()
		{
			return (IEnumerator<T>)((IEnumerable)this).GetEnumerator();
		}
		#endregion
	}
}