﻿using System;
using CoreLib.Io;

using MapUtility;

namespace MapEditor
{
	public static class RadishHelper
	{
		#region Public Interface
		/// <summary>
		/// If the stream is null, will make the necessary stream.
		/// </summary>
		/// <param name="x"></param>
		/// <param name="y"></param>
		/// <param name="blocks"></param>
		/// <param name="stream"></param>
		/// <returns></returns>
		public static ProtoBlock LoadBlockData(int x, int y, BlockHolder[,] blocks, IStreamReader stream)
		{
			BlockHolder rightBlock = blocks[x, y];
			if (rightBlock == null)
			{
				Console.WriteLine("Missing block! xOffset: {0} yOffset: {1}", x, y);
				Console.WriteLine("Press Enter to exit...");
				Console.ReadLine();
				return null;
			}
			ProtoBlock data;

			if (stream == null)
			{
				using (stream = new ByteReader(rightBlock.Path, false))
				{
					stream.Position = rightBlock.Position; //This is right after the length field
					data = ProtoBlock.Read(stream, rightBlock);
				}
			}
			else
			{
				stream.Position = rightBlock.Position; //This is right after the length field
				data = ProtoBlock.Read(stream, rightBlock);
			}
			data.TileData.HasTile = false;
			return data;
		}

		public static void MakeOtherRadishes()
		{
			RadishMaker.CombineRadishes();
			MakeLightmap2Radish();
			MakeLightmap3Radish();
			if (Settings.HasFourLightmaps)
				MakeLightmap4Radish();
		}
		public static void MakeLightmap2Radish()
		{
			MakeRadish(Settings.Lm2Width, Settings.Lm2Height, "lightmap2", 
				Shared.Lightmap1RadishHeaders, Shared.Lightmap1RadishPath, Shared.Lightmap2RadishPath, false);
		}
		public static void MakeLightmap3Radish()
		{
			MakeRadish(Settings.Lm3Width, Settings.Lm3Height, "lightmap3",
				Shared.Lightmap2RadishHeaders, Shared.Lightmap2RadishPath, Shared.Lightmap3RadishPath,
				false
				//Settings.HasFourLightmaps ? false : true
				);
		}
		public static void MakeLightmap4Radish()
		{
			MakeRadish(Settings.Lm4Width, Settings.Lm4Height, "lightmap4", 
				Shared.Lightmap3RadishHeaders, Shared.Lightmap3RadishPath, Shared.Lightmap4RadishPath, true);
		}
		#endregion

		#region Making Radish
		/// <summary>
		/// Source can be null if there are multiple sources.
		/// </summary>
		/// <param name="xlength"></param>
		/// <param name="ylength"></param>
		/// <param name="name"></param>
		/// <param name="blocks"></param>
		/// <param name="source"></param>
		/// <param name="destination"></param>
		private static void MakeRadish
			(int xlength, int ylength, string name, BlockHolder[,] blocks, string source, string destination, bool removeBorders)
		{
			Console.WriteLine("Making {0} radish", name);
			Console.Write("Doing column ");
			if (File.Exists(destination))
				System.IO.File.Delete(destination);
			using (BinaryStream headerStream = new BinaryStream(destination.Substring
				(0, destination.LastIndexOf('.')) + Shared.HeaderExtension))
			{
				using (BinaryStream writeStream = new BinaryStream(destination))
				{
					using (ByteReader readStream = new ByteReader(source))
					{
						MakeRadishHelper(xlength, ylength, blocks, readStream, writeStream, headerStream, removeBorders);
					}
				}
			}
			Console.WriteLine();
			Console.WriteLine("Done with {0} radish!", name);
		}
		private static void MakeRadishHelper
			(int xLength, int yLength, BlockHolder[,] blocks, IStreamReader readStream, 
			IStreamWriter dataStream, IStreamWriter headerStream, bool removeBorders)
		{
			for (int x = 0; x < xLength; x++) //Iteration of lightmap 2
			{
				string num = x.ToString();
				Console.Write(num);
				for (int y = 0; y < yLength; y++) //Iteration of lightmap 2
				{
					ProtoBlock topLeft = LoadBlockData(x * 2, y * 2, blocks, readStream);
					ProtoBlock topRight = LoadBlockData(x * 2 + 1, y * 2, blocks, readStream);
					ProtoBlock bottomLeft = LoadBlockData(x * 2, y * 2 + 1, blocks, readStream);
					ProtoBlock bottomRight = LoadBlockData(x * 2 + 1, y * 2 + 1, blocks, readStream);
					ProtoBlock newBlock = ShrinkBlocks(topLeft, topRight, bottomLeft, bottomRight);
					if (newBlock != null)
					{
						if (removeBorders)
						{
							newBlock.Ownership.RemoveBorders();
						}
						newBlock.Write(dataStream, headerStream);
					}
				}
				for (int i = 0; i < num.Length; i++)
					Console.Write('\b');
			}
		}
		#endregion

		#region Shrinking Blocks
		private static ProtoBlock ShrinkBlocks(ProtoBlock topLeft, ProtoBlock topRight, ProtoBlock bottomLeft, ProtoBlock bottomRight)
		{
			if (topLeft == null || topRight == null || bottomLeft == null || bottomRight == null)
				return null;

			ProtoBlock ret = new ProtoBlock();
			ret.XOffset = (short)(topLeft.XOffset / 2);
			ret.YOffset = (short)(topLeft.YOffset / 2);

			if (topLeft.Grayscale.Compressed && topRight.Grayscale.Compressed && 
				bottomLeft.Grayscale.Compressed && bottomRight.Grayscale.Compressed &&
				topLeft.Ownership.Compressed && topRight.Ownership.Compressed &&
				bottomLeft.Ownership.Compressed && bottomRight.Ownership.Compressed)
			{
				PixelOwnership common = topLeft.Ownership[0];
				byte grayscale = topLeft.Grayscale[0];
				if (topRight.Ownership[0] == common && bottomRight.Ownership[0] == common && bottomLeft.Ownership[0] == common &&
					topRight.Grayscale[0] == grayscale && bottomRight.Grayscale[0] == grayscale && bottomLeft.Grayscale[0] == grayscale)
				{
					ret.Grayscale.SetAll(grayscale);
					ret.Ownership.SetAll(common);
					return ret;
				}
			}

			ShrinkBlockPart(topLeft, ret, false, false);
			ShrinkBlockPart(topRight, ret, true, false);
			ShrinkBlockPart(bottomLeft, ret, false, true);
			ShrinkBlockPart(bottomRight, ret, true, true);
			return ret;
		}
		private static void ShrinkBlockPart(ProtoBlock toShrink, ProtoBlock final, bool xShifted, bool yShifted)
		{
			int xShift = xShifted ? 16 : 0;
			int yShift = yShifted ? 16 : 0;

			if (toShrink.Grayscale.Compressed && toShrink.Ownership.Compressed)
			{
				int xEnd = xShift + 16;
				int yEnd = yShift + 16;
				byte grayscale = toShrink.Grayscale[0];
				PixelOwnership ownership = toShrink.Ownership[0];
				for (int x = xShift; x < xEnd; x++)
				{
					for (int y = yShift; y < yEnd; y++)
					{
						final.Grayscale[x, y] = grayscale;
						final.Ownership[x, y] = ownership;
					}
				}
			}
			else
			{
				for (int x = 0; x < 16; x++)
				{
					for (int y = 0; y < 16; y++)
					{
						final.Grayscale[xShift + x, yShift + y] = ShrinkGrayscale(toShrink.Grayscale, x * 2, y * 2);
						final.Ownership[xShift + x, yShift + y] = ShrinkPixelOwnership(toShrink.Ownership, x * 2, y * 2);
					}
				}
			}
		}
		private static PixelOwnership ShrinkPixelOwnership(OwnershipData data, int xBegin, int yBegin)
		{
			PixelOwnership one = data[xBegin, yBegin];
			PixelOwnership two = data[xBegin + 1, yBegin];
			PixelOwnership three = data[xBegin, yBegin + 1];
			PixelOwnership four = data[xBegin + 1, yBegin + 1];

			//If at least three are equal we return that (This also covers cases where all are equal)
			if (
				(one == two && two == three) ||
				(one == three && three == four) ||
				(one == two && two == four)
				)
				return one;
			if (two == three && three == four)
				return two;

			//If two are equal and the other two are not equal to each other
			if (
				(one == two && three != four) ||
				(one == three && two != four) ||
				(one == four && two != three)
				)
				return one;
			if (
				(two == three && one != four) ||
				(two == four && one != three)
				)
				return two;
			if (three == four && one != two)
				return three;

			return one;

			/*
			//If two are the same and other two are also the same
			if (
				(one == two && three == four) ||
				(one == three && two == four) ||
				(one == four && two == three)
				)
				return one;

			//If they are all different
			return one; //Yes, this is redundant, but we might change things later so keep it
			*/
		}
		private static byte ShrinkGrayscale(ColoringData data, int xBegin, int yBegin)
		{
			byte one = data[xBegin, yBegin];
			byte two = data[xBegin + 1, yBegin];
			byte three = data[xBegin, yBegin + 1];
			byte four = data[xBegin + 1, yBegin + 1];
			return (byte)((one + two + three + four) / 4);
		}
		#endregion
	}
}