﻿using System;
using System.Collections.Generic;
using System.Drawing;
using CoreLib;
using CoreLib.Graphics;
using CoreLib.Graphics.Photoshop;
using CoreLib.Io;
using MapUtility;
using StreamWriter = System.IO.StreamWriter;
using System.Threading;
using System.Linq;

namespace MapEditor
{
	/// <summary>
	/// Inherit from this to make radishes from a file.
	/// </summary>
	public abstract class RadishMaker
	{
		#region Constant strings
		public const string Extension = ".radish";
		public const string FileSearchString = "*.radish";
		#endregion

		BinaryStream dataStream;
		BinaryStream headerStream;
		protected LineData FileData { get; private set; }

		public abstract FileTypes Handler { get; }

		#region Result Pics
		protected ImageData<RgbData> OwnershipPic { get; set; }
		protected ImageData<ArgbData> BorderPic { get; set; }
		protected ImageData<ArgbData> TilesPic { get; set; }
		protected ImageData<GrayscaleData> GrayscalePic { get; set; }
		#endregion

		protected abstract void LoadFile();
		private List<ushort> AdjanciableProvinces = new List<ushort>();
		protected RadishMaker(LineData data)
		{
			FileData = data;
			//Console.WriteLine("Working on {0}", FileData.FilePath);
			LoadFile();
			//Get the provinces that can be bordered (Presumably land)
			if (BorderPic != null)
			{
				for (int x = 0; x < BorderPic.Width; x++)
				{
					for (int y = 0; y < BorderPic.Height; y++)
					{
						if (BorderPic[x, y].A != 0)
						{
							ushort val = Global.ColorToId[OwnershipPic[x, y]];
							if (!AdjanciableProvinces.Contains(val))
								AdjanciableProvinces.Add(val);
						}
					}
				}
			}

			//Here we begin making the actual file
			string pathstr = data.FilePath.Split('\\')[data.FilePath.Split('\\').Length - 1];
			string path = Shared.RadishFolder + pathstr.Split('.')[pathstr.Split('.').Length - 2] + RadishMaker.Extension;
			if (System.IO.File.Exists(path))
				System.IO.File.Delete(path);
			using (headerStream = new BinaryStream(path.Substring(0, path.LastIndexOf('.')) + Shared.HeaderExtension))
			{
				using (dataStream = new BinaryStream(path, System.IO.FileMode.Create))
				{
					if (data.HasSpecialBaseLayer)
					{
						foreach (Point p in data.BaseLayerPoints)
							DoBlock(p);
					}
					else
					{
						for (int x = 0; x < OwnershipPic.Width; x += 32)
							for (int y = 0; y < OwnershipPic.Height; y += 32)
								DoBlock(new Point(x / 32, y / 32));
					}
				}
			}
		}
		private void DoBlockTiles(ProtoBlock b)
		{
			if (TilesPic == null) return;
			if (TilesPic.RegionIsTransparent(32 * (b.XOffset - FileData.xOffset), 32 * (b.YOffset - FileData.yOffset), 32, 32)) return;
			b.TileData.HasTile = true;
			ArgbData[] rectiles;
			ImageData<ArgbData>.Copy(TilesPic, 32 * (b.XOffset - FileData.xOffset),
				32 * (b.YOffset - FileData.yOffset), 32, 32, out rectiles);
			b.TileData.LoadFromData(rectiles);
		}
		private unsafe void DoBlockGrayscale(ProtoBlock b)
		{
			GrayscaleData[] stuff;
			ImageData<GrayscaleData>.Copy(GrayscalePic, b.XOffset * 32 - FileData.xOffset * 32,
				b.YOffset * 32 - FileData.yOffset * 32, 32, 32, out stuff);
			for (int i = 0; i < 1024; i++)
			{
				stuff[i] >>= 2;
			}

			fixed (GrayscaleData* p = &stuff[0])
			{
				b.Grayscale.LoadDataFrom((byte*)p, true);
			}
			//b.Grayscale.LoadDataFrom(Array.ConvertAll(stuff, color => color.Strength), true);
		}
		private void DoBlockOwnership(ProtoBlock b)
		{
			PixelOwnership po = new PixelOwnership();
			ArgbData borderPixel;
			for (int x = 0; x < 32; x++)
				for (int y = 0; y < 32; y++)
				{
					po.OwnerId = Global.ColorToId[OwnershipPic[x + (b.XOffset - FileData.xOffset) * 32,
						y + (b.YOffset - FileData.yOffset) * 32]];
					if (BorderPic != null)
					{
						borderPixel = BorderPic[x + (b.XOffset - FileData.xOffset) * 32,
							y + (b.YOffset - FileData.yOffset) * 32];

						if (borderPixel.A != 0) //We will figure it out later
						{
							BorderProvinces.TryAdd(po.OwnerId);
							po.IsBorder = true;
							if (borderPixel.B == 255)
								po.IsTypeOne = true;
							else //KEEP THIS!!!
								po.IsTypeOne = false;
						}
						else //KEEP THIS!!!!
							po.IsBorder = false;
					}
					b.Ownership[x, y] = po;
				}
			b.Ownership.TryToCompress();
		}

		private void DoBlock(Point p)
		{
			ProtoBlock b = new ProtoBlock();
			b.XOffset = (short)(FileData.xOffset + p.X);
			b.YOffset = (short)(FileData.yOffset + p.Y);

			DoBlockTiles(b);
			DoBlockGrayscale(b);
			DoBlockOwnership(b);

			b.Write(dataStream, headerStream);
		}

		/// <summary>
		/// Two purposes: 
		/// 1) Put all radishes into one file for easy access
		/// 2) Get bordering province IDs
		/// </summary>
		public static void CombineRadishes()
		{
			BorderProvinces.Load();
			string[] radishes = System.IO.Directory.GetFiles(Shared.RadishFolder, RadishMaker.FileSearchString);
			BlockHolder[,] blockHolder = Shared.LoadBlockHeaders(Settings.Lm1Width, Settings.Lm1Height, radishes);

			try
			{
				using (BinaryStream headerStream = new BinaryStream(Shared.Lightmap1RadishPath.Substring
					(0, Shared.Lightmap1RadishPath.LastIndexOf('.')) + Shared.HeaderExtension))
				{
					using (BinaryStream dataStream = new BinaryStream(Shared.Lightmap1RadishPath, System.IO.FileMode.Create))
					{
						Console.WriteLine("Doing lightmap1 radishes...");
						Console.Write("Doing column ");
						for (int x = 0; x < Settings.Lm1Width; x++)
						{
							string num = x.ToString();
							Console.Write(num);
							for (int y = 0; y < Settings.Lm1Height; y++)
							{
								#region Declare points
								Point leftLoc = new Point(x - 1, y);
								Point rightLoc = new Point(x + 1, y);
								Point topLoc = new Point(x, y - 1);
								Point bottomLoc = new Point(x, y + 1);
								#endregion
								#region Wrapping
								bool hasTop = y != 0;
								bool hasBottom = y != Settings.Lm1Height - 1;
								if (x == 0)
									leftLoc.X = Settings.Lm1Width - 1;
								if (x == Settings.Lm1Width - 1)
									rightLoc.X = 0;
								if (y == 0)
									hasTop = false;
								if (y == Settings.Lm1Height - 1)
									hasBottom = false;
								#endregion
								#region Load necessary streams
								EnsureBlockStreamLoaded(blockHolder[x, y]);
								EnsureBlockStreamLoaded(blockHolder[leftLoc.X, leftLoc.Y]);
								EnsureBlockStreamLoaded(blockHolder[rightLoc.X, rightLoc.Y]);
								if (hasTop)
									EnsureBlockStreamLoaded(blockHolder[topLoc.X, topLoc.Y]);
								if (hasBottom)
									EnsureBlockStreamLoaded(blockHolder[bottomLoc.X, bottomLoc.Y]);
								#endregion
								#region Loading blocks
								ProtoBlock current = RadishHelper.LoadBlockData(x, y, blockHolder, cachedStreams[blockHolder[x, y].Path]);
								ProtoBlock left = RadishHelper.LoadBlockData(leftLoc.X, leftLoc.Y, blockHolder,
									cachedStreams[blockHolder[leftLoc.X, leftLoc.Y].Path]);
								ProtoBlock right = RadishHelper.LoadBlockData(rightLoc.X, rightLoc.Y, blockHolder,
									cachedStreams[blockHolder[rightLoc.X, rightLoc.Y].Path]);
								ProtoBlock top;
								ProtoBlock bottom;
								if (hasTop)
									top = RadishHelper.LoadBlockData(topLoc.X, topLoc.Y, blockHolder,
										cachedStreams[blockHolder[topLoc.X, topLoc.Y].Path]);
								else
									top = null;
								if (hasBottom)
									bottom = RadishHelper.LoadBlockData(bottomLoc.X, bottomLoc.Y, blockHolder,
										cachedStreams[blockHolder[bottomLoc.X, bottomLoc.Y].Path]);
								else
									bottom = null;
								#endregion
								WriteData(dataStream, headerStream, current, left, right, top, bottom);
							}
							ConsoleEx.Backspace(num.Length);
						}
						Console.WriteLine("\b...and done!");
					}
				}
			}
			finally
			{
				foreach (ByteReader br in cachedStreams.Values)
					br.Dispose();
			}
		}

		private static List<Point> testPoints;
		private static void WriteData(IStreamWriter dataStream, IStreamWriter headerStream,
			ProtoBlock current, ProtoBlock left, ProtoBlock right, ProtoBlock top, ProtoBlock bottom)
		{
			const int Movement = 5;

			#region Initialize list of points
			if (testPoints == null)
			{
				Point toAdd = new Point();
				testPoints = new List<Point>(Movement * 4);
				for (int i = 0; i < Movement * 4; i++)
					testPoints.Add(toAdd);
			}
			#endregion

			if (!current.Ownership.Compressed) //Borders are certainly not 32 pixels wide
			{
				for (int x = 0; x < 32; x++)
				{
					for (int y = 0; y < 32; y++)
					{
						PixelOwnership po = current.Ownership[x, y];
						if (po.IsBorder)
						{
							#region FindPossiblePoints
							int counter = 0;
							Point pt = new Point();
							for (int i = 1; i < Movement + 1; i++)
							{
								pt.X = x + i;
								pt.Y = y;
								testPoints[counter++] = pt;
								pt.X = x - i;
								//pt.Y = y; //Reduntant, but keep this line to be clear
								testPoints[counter++] = pt;
								pt.X = x;
								pt.Y = y + i;
								testPoints[counter++] = pt;
								//pt.X = x; //Reduntant
								pt.Y = y - i;
								testPoints[counter++] = pt;
							}
							#endregion

							PixelOwnership other;
							foreach (Point p in testPoints)
							{
								other = GetPixelFromCoords(current, left, right, top, bottom, p.X, p.Y);
								if (other.OwnerId == po.OwnerId) continue;
								if (!BorderProvinces.Contains(other.OwnerId)) continue;
								po.OtherOwner = other.OwnerId;
								break;
							}
							if (po.OtherOwner == 0)
								po.IsBorder = false;
							current.Ownership[x, y] = po;
						}
					}
				}
			}
			current.Write(dataStream, headerStream);
		}

		private static PixelOwnership GetPixelFromCoords
			(ProtoBlock current, ProtoBlock left, ProtoBlock right, ProtoBlock top, ProtoBlock bottom, int x, int y)
		{
			if (x >= 0 && x < 32 && y >= 0 && y < 32)
			{
				//Nice and simple, is in the area
				return current.Ownership[x, y];
			}
			else if (x < 0)
			{
				return left.Ownership[32 + x, y];
			}
			else if (y < 0)
			{
				if (top != null)
					return top.Ownership[x, 32 + y];
				else
					return new PixelOwnership();
			}
			else if (x >= 32)
			{
				return right.Ownership[x - 32, y];
			}
			else // if (p.Y >= 32)
			{
				if (bottom != null)
					return bottom.Ownership[x, y - 32];
				else
					return new PixelOwnership();
			}
		}
		private static void EnsureBlockStreamLoaded(BlockHolder holder)
		{
			if (!cachedStreams.ContainsKey(holder.Path))
				cachedStreams.Add(holder.Path, new ByteReader(holder.Path, false));
		}
		private static Dictionary<string, ByteReader> cachedStreams = new Dictionary<string, ByteReader>(15);

		public static void MakeFiles(List<LineData> data)
		{
			Console.WriteLine("Working...");
			TaskQueue tq = new TaskQueue(data.Count);
			foreach (LineData ld in data)
			{
				tq.Add(new MakeFileTask(ld, Console.CursorTop, tq));
			}
			tq.StartWork();
			tq.FinishAll();
			Console.WriteLine("Done!");
		}

		public static void MakeFile(LineData data)
		{
			switch (data.FileType)
			{
				case SimpleImageRadishMaker.HandlerOf:
					new SimpleImageRadishMaker(data);
					break;
				case PsdRadishMaker.HandlerOf:
					new PsdRadishMaker(data);
					break;
			}
			BorderProvinces.Save();

		}
	}

	public sealed class SimpleImageRadishMaker : RadishMaker
	{
		public const FileTypes HandlerOf = FileTypes.SimpleImages;
		public override FileTypes Handler
		{
			get { return HandlerOf; }
		}

		public SimpleImageRadishMaker(LineData data) : base(data) { }

		protected override void LoadFile()
		{
			if (!LoadAPic(true, FileData.ProvinceOwnershipPath, ColorType.Rgb)) return;
			OwnershipPic = image.CastAs<RgbData>();
			if (!LoadAPic(true, FileData.GrayscalePath, ColorType.Grayscale)) return;
			GrayscalePic = image.CastAs<GrayscaleData>();
			if (LoadAPic(false, FileData.BordersPath, ColorType.Argb))
				BorderPic = image.CastAs<ArgbData>();
			if (LoadAPic(false, FileData.TilesPath, ColorType.Argb))
				TilesPic = image.CastAs<ArgbData>();
		}

		private static ImageData image;
		/// <summary>
		/// 
		/// </summary>
		/// <param name="required"></param>
		/// <param name="path"></param>
		/// <param name="layer"></param>
		/// <param name="type"></param>
		/// <returns>False if there was an error</returns>
		private static bool LoadAPic(bool required, string path, ColorType type)
		{
			if (!File.Exists(path))
			{
				if (required)
				{
					Console.WriteLine("File Missing: {0}", path);
					Console.WriteLine("Press any key...");
					Console.ReadKey();
				}
				return false;
			}
			Bitmap bmp = new Bitmap(path);

			switch (type)
			{
				case ColorType.Argb:
					image = ImageData<ArgbData>.ToData(bmp);
					break;
				case ColorType.Grayscale:
					image = ImageData<GrayscaleData>.ToData(bmp);
					break;
				case ColorType.Rgb:
					image = ImageData<RgbData>.ToData(bmp);
					break;
				default:
					throw new ArgumentOutOfRangeException("type");
			}
			return true;
		}

	}
	public sealed class PsdRadishMaker : RadishMaker
	{
		private FileReader reader;

		public PsdRadishMaker(LineData data) : base(data) { }

		private Layer LoadLayer(string name, ColorType type)
		{
			Layer toLoad = reader.Layers[name];
			if (toLoad == null)
			{
				Console.WriteLine("Layer " + name + " missing! Exiting");
				return null;
			}
			toLoad.LoadPic(type);
			return toLoad;
		}

		public const FileTypes HandlerOf = FileTypes.Photoshop;
		public override FileTypes Handler
		{
			get { return HandlerOf; }
		}

		protected override void LoadFile()
		{
			reader = new FileReader(FileData.FilePath);
			Layer ownership = LoadLayer("ProvID", ColorType.Rgb);
			Layer grayscale = LoadLayer("Colorscales", ColorType.Grayscale);
			OwnershipPic = (ImageData<RgbData>)ownership.PictureData;
			GrayscalePic = (ImageData<GrayscaleData>)grayscale.PictureData;

			//Load tiles
			Layer tiles = reader.Layers["Tiles"] as Layer;
			if (tiles != null)
			{
				tiles.LoadPic(ColorType.Argb);
				TilesPic = (ImageData<ArgbData>)tiles.PictureData;
			}

			//Border
			Layer border = reader.Layers["Borders"] as Layer;
			if (border != null)
			{
				border.LoadPic(ColorType.Argb);
				BorderPic = (ImageData<ArgbData>)border.PictureData;
			}
		}
	}

	class MakeFileTask : Task
	{
		private static object ConsoleLock = new object();

		public MakeFileTask(LineData data, int cursorRow, TaskQueue tq)
			: base("xxx-abc")
		{
			this.data = data;
			this.cursorRow = cursorRow;
			this.tq = tq;
		}

		private LineData data;
		private int cursorRow;
		private TaskQueue tq;

		public override void TaskOverride()
		{
			lock (ConsoleLock)
			{
				string[] statusData = tq.GetCurrentTaskDescriptions();
				int rowtogetto = cursorRow + statusData.Length + 2;
				Console.CursorTop = cursorRow;
				Console.CursorLeft = 0;
				while (Console.CursorTop < rowtogetto)
					Console.Write(' ');

				Console.CursorTop = cursorRow;
				Console.CursorLeft = 0;

				Console.WriteLine("Status as of " + DateTime.Now);
				for (int i = 0; i < statusData.Length; i++)
					Console.WriteLine("Thread {0}: {1}", i + 1, statusData[i]);

			}
			RadishMaker.MakeFile(data);
		}

		public override string Description
		{
			get { return "Doing " + data.FilePath; }
		}
	}
}