﻿using System;
using System.Diagnostics;
using CoreLib.Graphics;
using CoreLib.Io;

namespace MapEditor
{
	/// <summary>
	/// Manages the tile for a block
	/// </summary>
	public class Tile
	{
		#region Data
		private bool hasTile;
		private ArgbData[] tiles;
		public bool HasTile
		{
			[DebuggerHidden]
			get { return hasTile; }
			[DebuggerHidden]
			set
			{
				if (value)
				{
					if (tiles == null)
						tiles = new ArgbData[32 * 32];
				}
				else
					tiles = null;
				hasTile = value;
			}
		}
		#endregion

		#region Image editing
		[DebuggerHidden]
		public void SetPixel(int x, int y, ArgbData data)
		{
			tiles[y * 32 + x] = data;
		}
		#endregion

		#region Streaming
		public unsafe static Tile Read(IStreamReader stream)
		{
			Tile ret = new Tile();
			ret.HasTile = stream.ReadBool();
			if (ret.HasTile)
			{
				//for (int i = 0; i < 1024; i++)
				//	ret.tiles[i] = ArgbData.Read(stream);
				ArgbData[] vals = new ArgbData[1024];
				byte[] source = stream.ReadBytes(ArgbData.DataSize * 1024);
				fixed (byte* sourcePtr = &source[0])
				{
					fixed (ArgbData* destPtr = &vals[0])
					{
						CoreLib.Interop.Win32.Kernel.CopyMemory((IntPtr)destPtr, (IntPtr)sourcePtr, ArgbData.DataSize * 1024);
					}
					/*ArgbData* ptr = (ArgbData*)p;
					for (int i = 0; i < 1024; i++)
					{
						vals[i] = ptr[i];
					}*/
				}
				ret.tiles = vals;
			}
			return ret;
		}
		[Obsolete]
		public static Tile ReadStream(ByteReader stream)
		{
			Tile ret = new Tile();
			ret.HasTile = stream.ReadBool();
			if (ret.HasTile)
			{
				for (int i = 0; i < 1024; i++)
					ret.tiles[i] = ArgbData.Read(stream);
			}
			return ret;
		}
		[DebuggerHidden]
		public int WriteStream(IStreamWriter stream)
		{
			stream.WriteBool(HasTile);
			if (HasTile)
			{
				for (int i = 0; i < 1024; i++)
					tiles[i].Write(stream);
			}
			return HasTile ? 1024 * 4 + 1 : 1;
		}
		#endregion

		#region Constructors
		public Tile(ArgbData[] tiles)
		{
			LoadFromData(tiles);
		}
		public Tile() { }
		#endregion

		public void LoadFromData(ArgbData[] tiles)
		{
			if (tiles.Length != 1024)
				throw new ArgumentOutOfRangeException("tiles must have 1024 elements.");
			this.hasTile = true;
			this.tiles = tiles;
		}

		private static readonly ArgbData Green = new ArgbData(0, 255, 0);
		public ImageData<RgbData> MakeIntoImage()
		{
			if (!HasTile) return null;
			ImageData<RgbData> ret = new ImageData<RgbData>(32, 32);
			ArgbData data;
			for (int x = 0; x < 32; x++)
				for (int y = 0; y < 32; y++)
				{
					data = tiles[y * 32 + x];
					if (data.A == 0)
					{
						data = Green;
					}
					ret[x, y] = data;
				}
			return ret;
		}
	}
}