﻿using System;
using System.IO;
using CoreLib;
using CoreLib.Graphics;
using CoreLib.Io;
using MapUtility;
using System.Drawing;
using System.Drawing.Imaging;

namespace MapEditor
{
	public static class TileMaker
	{
		private static readonly int Width = Settings.Lm1Width;
		private static readonly int Height = Settings.Lm1Height;
		private const int BlocksPerTileRow = Global.TilesBlocksPerRow;
		private const int BlockLength = Global.BlockHeight;

		/// <summary>
		/// Looks in temp folder for radish files
		/// </summary>
		public unsafe static void MakeTiles()
		{
			string tilesPath = Settings.GamePath + @"\map\tiles.bmp";
			if (!CoreLib.Io.File.Exists(tilesPath))
			{
				Console.WriteLine("Did you set the GamePath in settings.txt?");
				Console.WriteLine(@"Does the following file exist: GamePath\map\tiles.bmp?");
				Console.WriteLine("The original tiles.bmp is required to copy the palette");
				Console.WriteLine("The tiles cannot be made.");
				return;
			}

			string[] files = Directory.GetFiles(Shared.RadishFolder, RadishMaker.FileSearchString);
			Console.Write("Scanning ");
			Tile[,] tiles = new Tile[Width, Height];
			int numTiles = 0;
			foreach (string s in files)
			{
				Console.Write(s);
				string headerPath = s.Substring(0, s.LastIndexOf('.')) + Shared.HeaderExtension;

				using (BinaryStream headerStream = new BinaryStream(headerPath, FileMode.Open))
				{
					using (BinaryStream dataStream = new BinaryStream(s, FileMode.Open))
					{
						while (headerStream.Position < headerStream.Length)
						{
							ProtoBlock t = ProtoBlock.Read(dataStream, headerStream);
							if (t.TileData != null && t.TileData.HasTile)
							{
								tiles[t.XOffset, t.YOffset] = t.TileData;
								numTiles++;
							}
						}
					}
				}
				ConsoleEx.Backspace(s.Length);
			}
			Console.WriteLine("\b...Done!");

			int tilesWidth = BlocksPerTileRow * BlockLength;
			int tilesHeight = (numTiles / BlocksPerTileRow) * BlockLength + BlockLength;

			ColorPalette tilesPalette = new Bitmap(tilesPath).Palette;
			Bitmap newTiles = new Bitmap(tilesWidth, tilesHeight, PixelFormat.Format8bppIndexed);
			newTiles.Palette = tilesPalette;
			NearestColorFinder finder = new NearestColorFinder(tilesPalette.Entries);


			BitmapData data = newTiles.LockBits(new Rectangle(0, 0, newTiles.Width, newTiles.Height), 
				ImageLockMode.ReadWrite, PixelFormat.Format8bppIndexed);

			byte* pixelStart = (byte*)data.Scan0;

			Console.Write("Creating files...");
			using (BinaryStream bw = new BinaryStream(Shared.IndexPath))
			{
				int iteration = 1;
				for (int x = 0; x < Width; x++)
					for (int y = 0; y < Height; y++)
					{
						int toWrite;
						if (tiles[x, y] == null)
							toWrite = 0;
						else
						{
							toWrite = iteration;
							int xShift = iteration % BlocksPerTileRow;
							int yShift = iteration / BlocksPerTileRow;
							xShift *= BlockLength;
							yShift *= BlockLength;
							Copy(xShift, yShift, pixelStart, tiles[x, y].MakeIntoImage(), finder);
							tiles[x, y] = null;
							iteration++;
						}

						if (Settings.IndexTblValuesAreFourBytes)
							bw.WriteInt32(toWrite);
						else
							bw.WriteUInt16((ushort)toWrite);
					}
			}
			newTiles.UnlockBits(data);
			newTiles.Save(Shared.OutputFolder + @"tiles.bmp", ImageFormat.Bmp);
	
			Console.WriteLine("Done!");
		}
		private static unsafe void Copy(int xStart, int yStart, byte* start, ImageData<RgbData> source, NearestColorFinder finder)
		{
			const int TilesBmpWidth = 32 * 256;

			for (int x = 0; x < source.Width; x++)
			{
				for (int y = 0; y < source.Height; y++)
				{
					start[(yStart + y) * TilesBmpWidth + xStart + x] = (byte)finder.GetNearestColorByIndex(source[x, y].ToColor());
				}
			}
		}
	}
}