﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.IO;
using System.Windows;
using System.Windows.Controls;
using CoreLib;
using CoreLib.Graphics;

namespace MapUtility
{
	/// <summary>
	/// Interaction logic for Window1.xaml
	/// </summary>
	public sealed partial class Display : Window, INotifyPropertyChanged
	{
		#region Options
		private int zoom;
		public int Zoom
		{
			get { return zoom; }
			set { zoom = value; OnPropertyChanged("Zoom"); }
		}
		private bool useColorscales;
		public bool UseColorscales
		{
			get { return useColorscales; }
			set
			{
				if (useColorscales != value)
				{
					useColorscales = value;
					OnPropertyChanged("useColorscales");
				}
			}
		}
		private bool showTiles;
		public bool ShowTiles
		{
			get { return showTiles; }
			set
			{
				if (showTiles != value)
				{
					showTiles = value;
					OnPropertyChanged("showTiles");
				}
			}
		}
		private bool showGrid;
		public bool ShowGrid
		{
			get { return showGrid; }
			set
			{
				if (showGrid != value)
				{
					showGrid = value;
					OnPropertyChanged("ShowGrid");
				}
			}
		}
		private List<Colorscales.Color> colorscales;
		public List<Colorscales.Color> Colorscales { get { return colorscales; } set { colorscales = value; } }
		public Colorscales.Color SelectedColor
		{
			get
			{
				return Lightmap.Scales;
			}
			set
			{
				if (Lightmap.Scales != value)
				{
					Lightmap.Scales = value;
					OnPropertyChanged("Colorscales");
					if (light != null && UseColorscales)
						canvas.Refresh();
					//if (light != null)
					//	viewer.Source = Lightmap.Convert(light.GetColoring<RgbData>(0, 0, 100, 100, Lightmap.ColorIndexColor));
				}
			}
		}
		private LightmapOption SelectedOption
		{
			get
			{
				int realZoom = zoom;
				if (!Settings.HasFourLightmaps)
					realZoom++;
				switch (realZoom)
				{
					case 1:
						return LightmapOption.LightMap4;
					case 2:
						return LightmapOption.LightMap3;
					case 3:
						return LightmapOption.LightMap2;
					case 4:
						return LightmapOption.LightMap1;
					default:
						throw new Exception("Lightmap zoom out of range");
				}
			}
		}
		#endregion

		#region Notify change
		public event PropertyChangedEventHandler PropertyChanged;
		private void OnPropertyChanged(string info)
		{
			if (PropertyChanged != null)
			{
				PropertyChanged(this, new PropertyChangedEventArgs(info));
			}
		}
		#endregion

		public Display()
		{
			Global.Log.WriteLine("In Display .ctor()");
			InitializeComponent();
			canvas = new System.Windows.Forms.Panel();
			Global.Log.WriteLine("Exiting Display .ctor()");

			host.Child = canvas;

			canvas.Paint += canvas_Paint;
			canvas.Width = canvas.Height = 100;
			canvas.MouseMove += canvas_MouseMove;
		}

		void canvas_MouseMove(object sender, System.Windows.Forms.MouseEventArgs e)
		{
			if (light == null)
				return;

			int x = -canvas.AutoScrollPosition.X + e.Location.X;
			int y = -canvas.AutoScrollPosition.Y + e.Location.Y;

			lblLocation.Text = string.Format("({0}, {1})", x.ToString(), y.ToString());
			lblBlkLoc.Text = string.Format("({0}, {1})", (x / 32).ToString(), (y / 32).ToString());
			MapBlock cur = light[x / 32, y / 32];
			cur.EnsureDecompressed();
			Leaf l = cur.FindCoordinates(x, y);
			lblOwnerId.Text = l.OwnerId.OwnerId.ToString();
			lblGrayscale.Text = l.GrayscaleColor.ToString();
			lblIsBorder.Text = l.OwnerId.IsBorder.ToString();
			if (l.OwnerId.IsBorder)
			{
				lblBorderInfo.Text = string.Format("IsTypeOne={0}, Other={1}",
					l.OwnerId.IsTypeOne.ToString(), l.OwnerId.OtherOwner.ToString());
				lblBorderInfo.Visibility = Visibility.Visible;
			}
			else
				lblBorderInfo.Visibility = Visibility.Collapsed;
			cur.DiscardData();
		}

		void host_KeyDown(object sender, System.Windows.Input.KeyEventArgs e)
		{
			const int move = 50;

			System.Drawing.Point proposed = new System.Drawing.Point(-canvas.AutoScrollPosition.X, -canvas.AutoScrollPosition.Y);

			if (e.Key == System.Windows.Input.Key.Down)
			{
				proposed.Y += move;
			}
			else if (e.Key == System.Windows.Input.Key.Up)
			{
				proposed.Y -= move;
			}
			else if (e.Key == System.Windows.Input.Key.Left)
			{
				proposed.X -= move;
			}
			else if (e.Key == System.Windows.Input.Key.Right)
			{
				proposed.X += move;
			}

			canvas.AutoScrollPosition = proposed;

			e.Handled = true;
		}		
		
		private Lightmap light;
		private System.Windows.Forms.Panel canvas;

		private void Button_Click(object sender, RoutedEventArgs e)
		{
			if (light == null || light.options != SelectedOption)
			{
				string path = Settings.GamePath + @"\map";
				Global.Log.Write("Loading {0} at {1}", SelectedOption, path);
				light = new Lightmap(SelectedOption, Settings.GamePath + @"\map");
				Global.Log.WriteLine(" ...And successful!");
			}
			light.LoadTileData();

			canvas.AutoScrollMinSize = new System.Drawing.Size(light.Width * 32, light.Height * 32);

			canvas.Refresh();
		}

		void canvas_Paint(object sender, System.Windows.Forms.PaintEventArgs e)
		{
			e.Graphics.TranslateTransform(canvas.AutoScrollPosition.X, canvas.AutoScrollPosition.Y);
			Rectangle rect = e.ClipRectangle;
			rect.X -= canvas.AutoScrollPosition.X;
			rect.Y -= canvas.AutoScrollPosition.Y;

			if (light != null)
			{
				if (UseColorscales)
				{
					light.DrawTrueColoredGrayscaleLayer(e.Graphics, rect);
				}
				else
				{
					light.Draw<RgbData>(e.Graphics, rect, Lightmap.ColorByProvinceId);
				}
			}

			UpdateGrid(e.Graphics, new Rectangle
				(-canvas.AutoScrollPosition.X, -canvas.AutoScrollPosition.Y, canvas.Width, canvas.Height));

			UpdateTiles(e.Graphics, rect);
		}

		private void ComboBox_SelectionChanged(object sender, SelectionChangedEventArgs e)
		{
			ComboBox cmb = sender as ComboBox;
			if (cmb.SelectedItem == null) return;
			SelectedColor = cmb.SelectedItem as Colorscales.Color;
		}

		private void Window_Loaded(object sender, RoutedEventArgs e)
		{
			if (!File.Exists(Settings.GamePath + @"\map\lightmap1.tbl"))
			{
				MessageBox.Show("Lightmap1.tbl was not found. Check that GamePath in settings.txt is correct.\r\n" +
					"The program will now exit.");
				Global.Log.WriteLine("Since lightmap1.tbl not found, now exiting. Searched path: "
					+ Settings.GamePath + @"\map\lightmap1.tbl");
				Global.Log.Flush();
				System.Diagnostics.Process.GetCurrentProcess().Kill();
			}
			Global.Log.Write("Loading colorscales...");
			#region Adding colorscales
			colorscales = new List<Colorscales.Color>();
			colorscales.Add(Lightmap.Colorscales.Plain);
			colorscales.Add(Lightmap.Colorscales.Black);
			colorscales.Add(Lightmap.Colorscales.Blue);
			colorscales.Add(Lightmap.Colorscales.Border);
			colorscales.Add(Lightmap.Colorscales.BorderOutline);
			colorscales.Add(Lightmap.Colorscales.Brown);
			colorscales.Add(Lightmap.Colorscales.DarkBlue);
			colorscales.Add(Lightmap.Colorscales.DarkBrown);
			colorscales.Add(Lightmap.Colorscales.DarkGray);
			colorscales.Add(Lightmap.Colorscales.DarkGreen);
			colorscales.Add(Lightmap.Colorscales.DarkOrange);
			colorscales.Add(Lightmap.Colorscales.DarkRed);
			colorscales.Add(Lightmap.Colorscales.DarkYellow);
			colorscales.Add(Lightmap.Colorscales.Gray);
			colorscales.Add(Lightmap.Colorscales.Green);
			colorscales.Add(Lightmap.Colorscales.LightBlue);
			colorscales.Add(Lightmap.Colorscales.LightBrown);
			colorscales.Add(Lightmap.Colorscales.LightGray);
			colorscales.Add(Lightmap.Colorscales.LightGreen);
			colorscales.Add(Lightmap.Colorscales.LightOrange);
			colorscales.Add(Lightmap.Colorscales.LightRed);
			colorscales.Add(Lightmap.Colorscales.LightYellow);
			colorscales.Add(Lightmap.Colorscales.Orange);
			colorscales.Add(Lightmap.Colorscales.Red);
			colorscales.Add(Lightmap.Colorscales.Water);
			colorscales.Add(Lightmap.Colorscales.White);
			colorscales.Add(Lightmap.Colorscales.Yellow);
			SelectedColor = Lightmap.Colorscales.Plain;
			#endregion
			Global.Log.WriteLine("And successful!");

			zoom = 1;
			this.DataContext = this;

			if (Settings.HasFourLightmaps)
			{
				sldZoom.Maximum = 4;
			}
			else
			{
				sldZoom.Maximum = 3;
			}

			this.Title = Shared.ProductName + " v" + AssemblyEx.GetVersion();
		}

		#region Drawing
		private static Font font = new Font("Times New Roman", 11);
		private static SolidBrush ColorBrush = new SolidBrush(System.Drawing.Color.Black);
		private static SolidBrush TransparentBrush = new SolidBrush(System.Drawing.Color.Transparent);
		private static System.Drawing.Pen LinePen = new System.Drawing.Pen(System.Drawing.Color.Black);

		private void UpdateGrid(Graphics g, Rectangle bounds)
		{
			if (light == null)
				return;
			if (chkShowGrid.IsChecked.Value)
			{
				int xmin = bounds.Left / 32 - 1; //Counted in blocks
				int xmax = bounds.Right / 32 + 1;
				int ymin = bounds.Top / 32 - 1;
				int ymax = bounds.Bottom / 32 + 1;
				if (xmin < 0) xmin = 0;
				if (xmax > light.Width) xmax = light.Width;
				if (ymin < 0) ymin = 0;
				if (ymax > light.Height) ymax = light.Height;

				for (int x = xmin; x < xmax; x++) //Vertical lines
				{
					g.DrawLine(LinePen, x * 32, ymin * 32, x * 32, ymax * 32);
				}
				for (int y = ymin; y < ymax; y++) //Horizontal lines
				{
					g.DrawLine(LinePen, xmin * 32, y * 32, xmax * 32, y * 32);
				}

				/*int writeLetterxbaseShift = xmin * 32;
				if (bounds.Left % 32 != 0)
					writeLetterxbaseShift += 32;
				int writeLetterybaseShift = ymin * 32;
				if (bounds.Top % 32 != 0)
					writeLetterybaseShift += 32;*/
				for (int x = xmin; x < xmax; x++) //Numbers on top
				{
					g.DrawString(x.ToString(), font, ColorBrush, x * 32, 0);
				}
				g.DrawString(ymin.ToString(), font, ColorBrush, 0, 16); //To avoid overlap
				for (int y = ymin + 1; y < ymax; y++) //Numbers on left
				{
					g.DrawString(y.ToString(), font, ColorBrush, 0, y * 32);
				}
			}
		}
		private void CheckBox_Checked(object sender, RoutedEventArgs e)
		{
			canvas.Refresh();
		}

		private void UpdateTiles(Graphics g, Rectangle bounds)
		{
			if (light == null)
				return;

			if (SelectedOption != LightmapOption.LightMap1)
				return;


			if (chkShowTiles.IsChecked.Value)
			{
				light.DrawTilesImage(g, bounds);
			}
		}

		private void chkShowTiles_Checked(object sender, RoutedEventArgs e)
		{
			canvas.Refresh();
		}
		#endregion

	}
}