﻿using System.Drawing;
using System.Linq;
using System.Windows;
using CoreLib.Compiliation.Ramixit;
using MapUtility;
using Point = System.Drawing.Point;

namespace ObjectEditor
{
	public sealed class ViewableObject
	{
		#region Data
		/// <summary>
		/// The path where the image of the item is located.
		/// </summary>
		public string Name { get; private set; }

		public int xColumn { get; private set; }
		public int yColumn { get; private set; }
		public int? IndexColumn { get; private set; }

		public Bitmap[] Images { get; private set; }
		public Point[] Sizes { get; private set; }
		public Point[] Offsets { get; private set; }
		#endregion

		public ViewableObject(Branch source)
		{
			Name = source.GetLeafValue("name");
			xColumn = int.Parse(source.GetLeafValue("xColumn"));
			yColumn = int.Parse(source.GetLeafValue("yColumn"));

			int indexColumn;
			if (int.TryParse(source.TryGetLeafValue("IndexColumn"), out indexColumn))
			{
				this.IndexColumn = indexColumn;
			}

			#region Load images
			Branch[] imgBranches = source.GetBranches("image").ToArray();
			Images = new Bitmap[imgBranches.Length];
			Sizes = new Point[imgBranches.Length];
			Offsets = new Point[imgBranches.Length];

			for (int i = 0; i < imgBranches.Length; i++)
			{
				DoImage(i, imgBranches[i]);
			}
			#endregion

			if (source.TryGetLeafValue("ViewAllAtOnce") == "yes")
			{
				Bitmap[] imgs = Images;
				Point[] sizes = Sizes;
				Point[] offsets = Offsets;

				Sizes = new Point[] { new Point(sizes.Max(pt => pt.X), sizes.Max(pt => pt.Y)) };
				Offsets = new Point[] { new Point(-Sizes[0].X / 2, -Sizes[0].Y / 2) };

				Bitmap final = new Bitmap(Sizes[0].X, Sizes[0].Y);

				using (Graphics g = Graphics.FromImage(final))
				{
					for (int i = 0; i < imgs.Length; i++)
					{
						g.DrawImageUnscaled(imgs[i], (Sizes[0].X - sizes[i].X) / 2, (Sizes[0].Y - sizes[i].Y) / 2);
					}
				}
				Images = new Bitmap[] { final };
			}
		}

		private void DoImage(int index, Branch imgBranch)
		{
			string path = Settings.GamePath + "\\" + imgBranch.GetLeafValue("path");

			if (!CoreLib.Io.File.Exists(path))
			{
				MessageBox.Show(string.Format("File {0} was not found! Now exiting", path), "Error");
				Application.Current.Shutdown();
				return;
			}

			Bitmap orig = new Bitmap(path);
			Branch boundBranch = imgBranch.TryGetBranch("bounds");
			int x, y, width, height;
			if (boundBranch != null)
			{
				x = int.Parse(boundBranch.GetLeafValue("x"));
				y = int.Parse(boundBranch.GetLeafValue("y"));
				width = int.Parse(boundBranch.GetLeafValue("width"));
				height = int.Parse(boundBranch.GetLeafValue("height"));
			}
			else
			{
				x = 0;
				y = 0;
				width = orig.Width;
				height = orig.Height;
			}

			Bitmap temp = new Bitmap(orig, width, height);
			for (int xPos = 0; xPos < width; xPos++)
			{
				for (int yPos = 0; yPos < height; yPos++)
				{
					temp.SetPixel(xPos, yPos, orig.GetPixel(x + xPos, y + yPos));
				}
			}

			Branch transparentBranch = imgBranch.GetBranch("transparency");
			var transparents = transparentBranch.GetBranches("color");
			foreach (Branch colorBranch in transparents)
			{
				temp.MakeTransparent
					(System.Drawing.Color.FromArgb(byte.Parse(colorBranch.GetLeafValue("R")),
					byte.Parse(colorBranch.GetLeafValue("G")), byte.Parse(colorBranch.GetLeafValue("B"))));
			}
			this.Images[index] = temp;
			this.Sizes[index] = new Point(temp.Width, temp.Height);

			Offsets[index] = new System.Drawing.Point(-this.Images[index].Width / 2, -this.Images[index].Height / 2);
		}
	}
}