﻿using System.Collections;
using System.Collections.Generic;
using CoreLib.Io;

namespace MapUtility
{
	public sealed class AdjacentTable
	{
		#region Fields
		private List<AdjacentProvinceInfo> Provinces = new List<AdjacentProvinceInfo>(3000);
		#endregion

		#region Constructors
		public AdjacentTable(string str)
		{
			List<int> offsets = new List<int>(3000);
			using (ByteReader stream = new ByteReader(str, true))
			{
				/* Offsets
				 * The first one is always 0
				 * The next are pairs specifying the beginning and end of a block
				 * The ending of one is the beginning of another
				 * A value must be multiplied by 12 for the byte offset
				 * Offsets are calculated from the end of the offsets
				 */
				int prevNumber = 0;
				int readNumber;
				while ((readNumber = stream.ReadInt32()) >= prevNumber)
				{
					offsets.Add(readNumber);
					prevNumber = readNumber;
				}
				stream.Position -= 4; //Shift back due to above
				/* Body
				 * Is made of blocks of 12 bytes
				 * The first four bytes is a connected province
				 * //The next four is 1 if there is a river in between, otherwise 0
				 * The next four appear to always be zero, something about rivers changed
				 * The next four is always 0xCDCD, there appears to be no special meaning
				 */
				for (int blocknum = 0; blocknum < offsets.Count - 1; blocknum++)
				{
					int currentLocation = 0;
					int length = offsets[blocknum + 1] - offsets[blocknum];
					Provinces.Add(new AdjacentProvinceInfo(blocknum));
					while (currentLocation < length)
					{
						ushort connectedProv = (ushort)stream.ReadInt32();
						//int river = stream.ReadInt32();
						//int nonsense = stream.ReadInt32();
						//if (river != 0)
						//	river.ToString();
						stream.Position += 8;
						currentLocation++;
						Provinces[blocknum].AddConnection(connectedProv);
					}
					Provinces[blocknum].Sort();
				}
			}
		}
		#endregion

		#region Data accessors
		public AdjacentProvinceInfo this[ushort provID]
		{
			get
			{
				return Provinces.Find(item => item.Id == provID);
			}
		}
		public ushort this[ushort provID, int index]
		{
			get
			{
				return this[provID][index];
			}
		}
		public int NumProvinces { get { return Provinces.Count; } }
		#endregion
	}
	public sealed class AdjacentProvinceInfo : IEnumerable<ushort>
	{
		#region Data
		/// <summary>
		/// A list of provinces connected to this.
		/// </summary>
		private List<ushort> ConnectedProvinces = new List<ushort>();
		/// <summary>
		/// The province ID of this AdjacentProvinceInfo.
		/// </summary>
		public int Id { get; private set; }
		#endregion

		#region Constructor
		public AdjacentProvinceInfo(int id)
		{
			this.Id = id;
		}
		#endregion

		#region List operations
		public void AddConnection(ushort connectedProvince)
		{
			if (!ConnectedProvinces.Contains(connectedProvince))
				ConnectedProvinces.Add(connectedProvince);
		}
		public void Sort()
		{
			ConnectedProvinces.Sort((x, y) => x - y);
		}
		public int IndexOf(ushort province)
		{
			return ConnectedProvinces.IndexOf(province);
		}
		public bool Contains(ushort connection)
		{
			return ConnectedProvinces.Contains(connection);
		}
		#endregion

		#region Data access
		public ushort this[int index]
		{
			get
			{
				return ConnectedProvinces[index];
			}
		}
		public int NumConnections { get { return ConnectedProvinces.Count; } }
		#endregion

		#region IEnumerable<ushort> Implementation
		public IEnumerator<ushort> GetEnumerator()
		{
			return ConnectedProvinces.GetEnumerator();

		}
		IEnumerator IEnumerable.GetEnumerator()
		{
			return ConnectedProvinces.GetEnumerator();
		}
		#endregion
	}
}