﻿using System;
using System.Collections.Generic;

namespace MapUtility
{
	public class Leaf : IEquatable<Leaf>
	{
		#region Basic properties
		/// <summary>
		/// 32 - 5
		/// 16 - 4
		/// 8  - 3
		/// 4  - 2
		/// 2  - 1
		/// 1  - 0
		/// </summary>
		public byte Level { get; private set; }
		public PixelOwnership OwnerId { get; set; }
		public byte GrayscaleColor { get; set; } //0 - 63
		public byte DeepGrayscaleColor { get { return (byte)(GrayscaleColor << 2); } }
		public int Shift { get { return 1 << Level; } }

		public ushort xCoordinate { get; private set; }
		public ushort yCoordinate { get; private set; }
		#endregion

		protected Leaf parent;

		#region Children
		public Leaf[] Children;
		public bool HasChildren { get { return Children != null; } }
		public Leaf TopLeftChild
		{
			get
			{
				if (Children == null)
					return null;
				return Children[3];
			}
		}
		public Leaf BottomLeftChild
		{
			get
			{
				if (Children == null)
					return null;
				return Children[1];
			}
		}
		public Leaf TopRightChild
		{
			get
			{
				if (Children == null)
					return null;
				return Children[2];
			}
		}
		public Leaf BottomRightChild
		{
			get
			{
				if (Children == null)
					return null;
				return Children[0];
			}
		}
		public void AddChild(Leaf child)
		{
			if (Children == null)
				Children = new Leaf[4];
			for (int i = 0; i < 4; i++)
			{
				if (Children[i] == null)
				{
					Children[i] = child;
					return;
				}
			}
		}
		#endregion

		public void SetAllOwners(PixelOwnership value)
		{
			this.OwnerId = value;
			if (this.HasChildren)
				for (int i = 0; i < 4; i++)
					this.Children[i].SetAllOwners(value);
		}

		#region Constructors
		public Leaf(Leaf parent, int position)
		{
			this.parent = parent;
			this.Level = (byte)(parent.Level - 1);
			this.xCoordinate = parent.xCoordinate;
			this.yCoordinate = parent.yCoordinate;
			if (position == 0 || position == 1)
				this.yCoordinate += (byte)this.Shift;
			if (position == 0 || position == 2)
				this.xCoordinate += (byte)this.Shift;
		}
		/// <summary>
		/// To help inheritors.
		/// </summary>
		protected Leaf(byte level, ushort xPos, ushort yPos) { this.Level = level; this.xCoordinate = xPos; this.yCoordinate = yPos; }
		#endregion

		public MapBlock TopParent
		{
			get
			{
				Leaf current = this;
				while (current.parent != null)
				{
					current = current.parent;
				}
				return (MapBlock)current;
			}
		}

		public override string ToString()
		{
			return string.Format("Level={0} OwnerId={1} GrayscaleColor={2} HasChildren={3} X={4} Y={5}",
				Level, OwnerId, GrayscaleColor, HasChildren ? "T" : "F", xCoordinate.ToString(), yCoordinate.ToString());
		}

		#region Neighbors
		public Leaf RightNeighbor
		{
			get
			{
				return FindCoordinates(this.xCoordinate + this.Shift, this.yCoordinate);
			}
		}
		public Leaf BottomNeighbor
		{
			get
			{
				return FindCoordinates(this.xCoordinate, this.yCoordinate + this.Shift);
			}
		}
		public Leaf BottomRightNeighbor
		{
			get
			{
				return FindCoordinates(this.xCoordinate + this.Shift, this.yCoordinate + this.Shift);
			}
		}

		private Leaf FindSolutionWayOne(int x, int y, int rightX, int? bottomY, MapBlock same)
		{
			Predicate<Leaf> action = l =>
				l.xCoordinate <= x
				&& l.xCoordinate >= x + l.Shift
				&& l.yCoordinate <= y
				&& l.yCoordinate >= y + l.Shift
				//l.xCoordinate == x && l.yCoordinate == y 
				//&& l.Level == this.Level
				;
			Leaf solution = same.FindLeaf(action);
			if (solution != null)
				return solution;
			solution = same.WholeMap[rightX, same.Y].FindLeaf(action);
			if (solution != null)
				return solution;
			if (bottomY != null)
			{
				solution = same.WholeMap[same.X, bottomY.Value].FindLeaf(action);
				if (solution != null)
					return solution;
				solution = same.WholeMap[rightX, bottomY.Value].FindLeaf(action);
				if (solution != null)
					return solution;
			}
			return null;
		}
		private Leaf FindSolutionWayTwo(int x, int y, int rightX, int? bottomY, MapBlock same)
		{
			Predicate<Leaf> action = l =>
				l.xCoordinate >= x
					//&& l.xCoordinate <= x + l.Shift
				&& l.yCoordinate >= y
				//&& l.yCoordinate <= y + l.Shift
				//l.xCoordinate == x && l.yCoordinate == y 
				//&& l.Level == this.Level
				;
			Leaf solution = same.FindLeaf(action);
			if (solution != null)
				return solution;
			solution = same.WholeMap[rightX, same.Y].FindLeaf(action);
			if (solution != null)
				return solution;
			if (bottomY != null)
			{
				solution = same.WholeMap[same.X, bottomY.Value].FindLeaf(action);
				if (solution != null)
					return solution;
				solution = same.WholeMap[rightX, bottomY.Value].FindLeaf(action);
				if (solution != null)
					return solution;
			}
			return null;
		}
		private Leaf FindSolution(int x, int y, int rightX, int? bottomY, MapBlock same, Predicate<Leaf> check)
		{
			Leaf solution = same.FindLeaf(check);
			if (solution != null)
				return solution;
			solution = same.WholeMap[rightX, same.Y].FindLeaf(check);
			if (solution != null)
				return solution;
			if (bottomY != null)
			{
				solution = same.WholeMap[same.X, bottomY.Value].FindLeaf(check);
				if (solution != null)
					return solution;
				solution = same.WholeMap[rightX, bottomY.Value].FindLeaf(check);
				if (solution != null)
					return solution;
			}
			return null;
		}
		public Leaf FindCoordinates(int x, int y)
		{

			MapBlock same = TopParent;

			if (y >= same.WholeMap.Height * 32)
				return null;
			if (x >= same.WholeMap.Width * 32)
				x -= same.WholeMap.Width * 32;

			int rightX = same.X + 1;
			if (rightX >= same.WholeMap.Width)
				rightX = 0;
			int? bottomY = same.Y + 1;
			if (bottomY >= same.WholeMap.Height)
				bottomY = null;

			Leaf wayOne = FindSolution(x, y, rightX, bottomY, same,
				l =>
				l.xCoordinate <= x
				&& l.xCoordinate + l.Shift - 1 >= x
				&& l.yCoordinate <= y
				&& l.yCoordinate + l.Shift - 1 >= y
				//l.xCoordinate == x && l.yCoordinate == y 
				//&& l.Level == this.Level
				);
			/*Leaf wayTwo = FindSolution(x, y, rightX, bottomY, same,
				l =>
				l.xCoordinate >= x
					//&& l.xCoordinate <= x + l.Shift
				&& l.yCoordinate >= y
				//&& l.yCoordinate <= y + l.Shift
				//l.xCoordinate == x && l.yCoordinate == y 
				//&& l.Level == this.Level
				);

			if (wayOne != wayTwo)
			{
				wayTwo.ToString();
			}*/

			return wayOne;
		}
		#endregion

		#region Visiting leaves
		public void DoAction(Action<Leaf> action)
		{
			action(this);
			if (HasChildren)
			{
				for (int i = 0; i < 4; i++)
				{
					this.Children[i].DoAction(action);
				}
			}
		}
		public Leaf FindLeaf(Predicate<Leaf> comparison)
		{
			if (!HasChildren && comparison(this)) //Things without children don't count
				return this;
			if (HasChildren)
			{
				for (int i = 3; i > -1; i--)
				{
					Leaf found = this.Children[i].FindLeaf(comparison);
					if (found != null)
						return found;
				}
			}
			return null;
		}
		#endregion

		#region Writing to data
		protected void AppendNumbers(List<bool> source)
		{
			source.Add(this.HasChildren);
			if (this.HasChildren && this.Level > 1)
			{
				for (int i = 0; i < 4; i++)
				{
					this.Children[i].AppendNumbers(source);
				}
			}
		}
		protected void AppendOwnerIdInfo(List<PixelOwnership> provinces, List<int> toAdd)
		{
			if (this.HasChildren)
			{
				for (int i = 0; i < 4; i++)
				{
					this.Children[i].AppendOwnerIdInfo(provinces, toAdd);
				}
			}
			else
			{
				toAdd.Add(provinces.IndexOf(this.OwnerId));
			}
		}
		protected void AppendColorInfo(List<byte> toAdd)
		{
			if (this.HasChildren)
			{
				for (int i = 0; i < 4; i++)
				{
					this.Children[i].AppendColorInfo(toAdd);
				}
			}
			else
			{
				toAdd.Add(this.GrayscaleColor);
			}
		}
		#endregion

		#region IEquatable<Leaf> Members

		public bool Equals(Leaf other)
		{
			if (this.HasChildren && other.HasChildren)
			{
				for (int i = 0; i < 4; i++)
				{
					if (!this.Children[i].Equals(other.Children[i]))
						return false;
				}
				return true;
			}
			if (!this.HasChildren && !other.HasChildren)
			{
				if (this.Level == other.Level
					&& this.OwnerId == other.OwnerId
					&& this.GrayscaleColor == other.GrayscaleColor)
					return true;
				return false; // TODO: Optimize when errors all gone
			}

			return false;
		}

		#endregion
	}
}