﻿//#define TRY

using System;
using System.Drawing;
using System.IO;
using CoreLib.Graphics;
using CoreLib.Io;
using System.Drawing.Imaging;

namespace MapUtility
{
	public class Lightmap : IEquatable<Lightmap>
	{
		#region Map Size
		public int Width; //In blocks
		public int Height;
		public int NumBlocks { get { return Height * Width; } }
		public int NumOffsets { get { return NumBlocks + 1; } }
		public LightmapOption options { get; private set; }
		#endregion

		private string path;

		#region Mapblocks
		MapBlock[] Blocks;
		public MapBlock this[int x, int y]
		{
			get
			{
				return Blocks[y * Width + x];
			}
			set
			{
				Blocks[y * Width + x] = value;
			}
		}
		#endregion

		#region Constructors
		public Lightmap(LightmapOption lightmap, string folder)
		{
			#region Load basic properties
			options = lightmap;
			switch (lightmap)
			{
				case LightmapOption.LightMap1:
					Width = Settings.Lm1Width;
					Height = Settings.Lm1Height;
					path = folder + @"\lightmap1.tbl";
					break;
				case LightmapOption.LightMap2:
					Width = Settings.Lm2Width;
					Height = Settings.Lm2Height;
					path = folder + @"\lightmap2.tbl";
					break;
				case LightmapOption.LightMap3:
					Width = Settings.Lm3Width;
					Height = Settings.Lm3Height;
					path = folder + @"\lightmap3.tbl";
					break;
				default: //LightMap4
					Width = Settings.Lm4Width;
					Height = Settings.Lm4Height;
					path = folder + @"\lightmap4.tbl";
					break;
			}
			#endregion


			Blocks = new MapBlock[NumBlocks];

			using (BinaryReader br = new BinaryReader(new FileStream(path, FileMode.Open, FileAccess.Read)))
			{
				int[] offsets;
				int OffsetWidth;
				if (Settings.LmOffsetsAreFourBytes)
				{
					offsets = Read4ByteOffsets(br);
					OffsetWidth = 4;
				}
				else //Offsets are 3 bytes wide
				{
					offsets = Read3ByteOffsets(br);
					OffsetWidth = 3;
				}

				int offsetsSectionEnd = NumOffsets * OffsetWidth;
				if (!Settings.LmOffsetsAreFourBytes)
				{
					while (offsetsSectionEnd % 2 != 0)
						offsetsSectionEnd++;

				}

				for (int y = 0; y < Height; y++)
				{
					for (int x = 0; x < Width; x++)
					{
						int length = offsets[y * Width + x + 1] - offsets[y * Width + x];
						br.BaseStream.Position = offsets[y * Width + x] + offsetsSectionEnd;// -2;
						Blocks[y * Width + x] = new MapBlock((ushort)x, (ushort)y, this, br.ReadBytes(length));
					}
				}
			}
		}

		private int[] Read4ByteOffsets(BinaryReader br)
		{
			int[] offsets = new int[NumOffsets];
			for (int i = 0; i < NumOffsets; i++)
			{
				offsets[i] = br.ReadInt32();
			}
			br.BaseStream.Position = NumOffsets * 4;
			return offsets;
		}
		private int[] Read3ByteOffsets(BinaryReader br)
		{
			const int AndVal = 255 | (255 << 8) | (255 << 16);

			int[] offsets = new int[NumOffsets];
			for (int i = 0; i < NumOffsets; i++)
			{
				offsets[i] = br.ReadInt32() & AndVal;
				br.BaseStream.Position--;
			}
			br.BaseStream.Position = NumOffsets * 3;
			return offsets;
		}
		public Lightmap(int width, int height, string path)
		{
			this.Width = width;
			this.Height = height;
			this.path = path;

			Blocks = new MapBlock[NumBlocks];
		}
		#endregion

		#region Colorscales
		public static Colorscales.Color Scales { get; set; }
		public static Colorscales Colorscales = new Colorscales();
		#endregion

		#region Drawing

		#region Generic coloring
		public ImageData<T> GetColoring<T>(int x, int y, int captureWidth, int captureHeight, Func<Leaf, T> colorer)
			where T : struct, IColorData
		{
			return GetColoring(x, y, captureWidth, captureHeight, colorer, true);
		}
		public ImageData<T> GetColoring<T>(int x, int y, int captureWidth, int captureHeight, Func<Leaf, T> colorer, bool discardData)
			where T : struct, IColorData
		{
			if (x + captureWidth > Width)
				captureWidth = Width - x;
			if (y + captureHeight > Height)
				captureHeight = Height - y;
			if (captureWidth <= 0 || captureHeight <= 0)
				return null;

			ImageData<T> ret = new ImageData<T>(captureWidth * 32, captureHeight * 32);
			for (int xPos = 0; xPos < captureWidth; xPos++)
			{
				for (int yPos = 0; yPos < captureHeight; yPos++)
				{
					int realX = x + xPos;
					int realY = y + yPos;
					Blocks[realY * Width + realX].EnsureDecompressed();
					DrawGenericLeafData(xPos * 32, yPos * 32, Blocks[realY * Width + realX], colorer, ret);
					if (discardData)
						Blocks[realY * Width + realX].DiscardData();
				}
			}
			return ret;
		}
		private void DrawGenericLeafData<T>(int x, int y, Leaf l, Func<Leaf, T> colorer, ImageData<T> canvas) 
			where T : struct, IColorData
		{
			if (l.HasChildren)
			{
				for (int i = 0; i < 4; i++)
				{
					Point origin = new Point(x + (i == 2 || i == 0 ? l.Shift >> 1 : 0), y + ((i == 1 || i == 0) ? l.Shift >> 1 : 0));
					DrawGenericLeafData(origin.X, origin.Y, l.Children[i], colorer, canvas);
				}
			}
			else
			{
				T val = colorer(l);
				canvas.FillRectangle(x, y, l.Shift, l.Shift, val);
			}
		}
		public void Draw<T>(Graphics g, Rectangle bounds, Func<Leaf, T> colorer) where T : struct, IColorData
		{
			Draw(g, bounds, colorer, 0, 0);
		}
		public void Draw<T>(Graphics g, Rectangle bounds, Func<Leaf, T> colorer, int xStart, int yStart) where T : struct, IColorData
		{
			int xmax = bounds.Right / 32 + 1;
			int ymax = bounds.Bottom / 32 + 1;
			xmax += xStart;
			ymax += yStart;
			if (xmax > Width) xmax = Width;
			if (ymax > Height) ymax = Height;

			Draw<T>(g, bounds, colorer, xStart, yStart, xmax, ymax);
		}
		public void Draw<T>(Graphics g, Rectangle bounds, Func<Leaf, T> colorer, int xStart, int yStart, int xMax, int yMax)
			where T : struct, IColorData
		{
			int xmin = bounds.Left / 32 - 1; //Counted in blocks
			int xmax = bounds.Right / 32 + 1;
			int ymin = bounds.Top / 32 - 1;
			int ymax = bounds.Bottom / 32 + 1;
			xmin += xStart;
			xmax += xStart;
			ymin += yStart;
			ymax += yStart;
			if (xmin < 0) xmin = 0;
			if (xmax > Width) xmax = Width;
			if (ymin < 0) ymin = 0;
			if (ymax > Height) ymax = Height;
			if (xMax > xmax) xMax = xmax;
			if (yMax > ymax) yMax = ymax;

			ImageData<T> temp = new ImageData<T>(32 * (xmax - xmin), 32 * (ymax - ymin));
			for (int y = ymin; y < yMax; y++)
			{
				for (int x = xmin; x < xMax; x++)
				{
					this[x, y].EnsureDecompressed();
					DrawGenericLeafData((x - xmin) * 32, (y - ymin) * 32, this[x, y], colorer, temp);
					this[x, y].DiscardData();
				}
			}
			g.DrawImageUnscaled(temp.InternalBitmap, (xmin - xStart) * 32, (ymin - yStart) * 32);
		}
		#endregion

		#region Border layer for images
		/// <summary>
		/// Note: Can return null.
		/// </summary>
		/// <returns></returns>
		public ImageData<ArgbData> GetBorderLayer(int x, int y, int width, int height)
		{
			if (x + width > Width)
				width = Width - x;
			if (y + height > Height)
				height = Height - y;
			if (width <= 0 || height <= 0)
				return null;

			bool NoBorders = true;
			//Check if are any borders
			for (int xPos = x; xPos < x + width; xPos++)
			{
				for (int yPos = y; yPos < y + height; yPos++)
				{
					Blocks[yPos * Width + xPos].EnsureDecompressed();
					if (!Blocks[yPos * Width + xPos].ForAllProvinceIds(id => !id.IsBorder))
					{
						NoBorders = false;
						goto done;
					}
				}
			}
		done:
			if (NoBorders)
			{
				for (int xPos = x; xPos < x + width; xPos++)
				{
					for (int yPos = y; yPos < y + height; yPos++)
					{
						Blocks[yPos * Width + xPos].DiscardData();
					}
				}
				return null;
			}

			return GetColoring<ArgbData>(x, y, width, height, delegate(Leaf l)
			{
				if (!l.OwnerId.IsBorder)
					return new ArgbData(0, 255, 255, 255);
				return l.OwnerId.IsTypeOne ? RgbData.Blue : RgbData.Red;
			});
		}
		#endregion

		#region Func<Leaf, T> Colorers where T : IColorData
		public static RgbData ColorIndexColor(Leaf l)
		{
			return Scales[ColorBySimpleGrayscale(l) >> 2];
		}
		public static RgbData ColorByProvinceId(Leaf l)
		{
			return Global.IdToColor[l.OwnerId.OwnerId];
		}
		public static GrayscaleData ColorBySimpleGrayscale(Leaf l)
		{
			if (l.Level == 0)
				return l.DeepGrayscaleColor;
			Leaf right = l.RightNeighbor;
			Leaf bottom = l.BottomNeighbor;
			Leaf bottomRight = l.BottomRightNeighbor;
			if (right == null || bottom == null || bottomRight == null)
				return l.DeepGrayscaleColor;
			if (right.GrayscaleColor == bottom.GrayscaleColor
				&& bottom.GrayscaleColor == bottomRight.GrayscaleColor
				&& bottomRight.GrayscaleColor == l.GrayscaleColor)
				return l.DeepGrayscaleColor;
			return new GrayscaleData(0);
		}
		#endregion

		#region Grayscale layer/colored grayscale layer
		private static void DrawLeaf(int x, int y, Leaf l, ImageData<RgbData> canvas)
		{
			#region If has children
			if (l.HasChildren)
			{
				int shift = l.Shift >> 1;
				DrawLeaf(x, y, l.TopLeftChild, canvas);
				DrawLeaf(x + shift, y, l.TopRightChild, canvas);
				DrawLeaf(x, y + shift, l.BottomLeftChild, canvas);
				DrawLeaf(x + shift, y + shift, l.BottomRightChild, canvas);
				return;
			}
			#endregion
			#region If is single pixel
			if (l.Level == 0)
			{
				canvas[x, y] = Scales[l.GrayscaleColor];
				return;
			}
			#endregion

			#region Declare neighbor leaves
			Leaf bottomRightLeaf = l.BottomRightNeighbor;
			Leaf bottomLeaf = l.BottomNeighbor;
			Leaf rightLeaf = l.RightNeighbor;
			#endregion

			#region If some neighbors are missing
			if (bottomRightLeaf == null || bottomLeaf == null || rightLeaf == null)
			{
				canvas.FillRectangle(x, y, l.Shift, l.Shift, Scales[l.GrayscaleColor]);
				return;
			}
			#endregion

			#region Declare grayscale values
			byte topLeft = l.DeepGrayscaleColor;
			byte right = rightLeaf.DeepGrayscaleColor;
			byte bottom = bottomLeaf.DeepGrayscaleColor;
			byte bottomRight = bottomRightLeaf.DeepGrayscaleColor;
			#endregion

			#region If all values are same
			if (topLeft == right && right == bottom && bottom == bottomRight) //Since all are same, simple filling in
			{
				canvas.FillRectangle(x, y, l.Shift, l.Shift, Scales[topLeft >> 2]);
				return;
			}
			#endregion

			byte[,] data = new byte[2, l.Shift + 1];
			data[0, 0] = topLeft;
			data[0, l.Shift] = bottom;
			data[1, l.Shift] = bottomRight;
			data[1, 0] = right;

			//int realMaxDiff = MaxDiff;
			//Is int because shifts can be negative, and sbytes might not have enough range
			int firstShift = bottom - topLeft;
			int secondShift = bottomRight - right;
			for (int i2 = 1; i2 < l.Shift; i2++)
			{
				data[0, i2] = (byte)(topLeft + (firstShift * i2 + 0.5) / l.Shift);
				data[1, i2] = (byte)(right + (secondShift * i2 + 0.5) / l.Shift);
			}

			for (int row = 0; row < l.Shift; row++)
			{
				int valShift = data[1, row] - data[0, row];
				for (int i2 = 0; i2 < l.Shift; i2++)
				{
					canvas[x + i2, y + row] = Scales[(byte)(data[0, row] + (valShift * i2 + 0.5) / l.Shift) >> 2];
				}
			}

		}
		public ImageData<RgbData> GetTrueColoredGrayscaleLayer(int x, int y, int captureWidth, int captureHeight)
		{
			if (x + captureWidth > Width)
				captureWidth = Width - x;
			if (y + captureHeight > Height)
				captureHeight = Height - y;
			if (captureWidth <= 0 || captureHeight <= 0)
				return null;

			ImageData<RgbData> ret = new ImageData<RgbData>(captureWidth * 32, captureHeight * 32);

			for (int xPos = 0; xPos < captureWidth; xPos++)
			{
				for (int yPos = 0; yPos < captureHeight; yPos++)
				{
					int realX = x + xPos;
					int realY = y + yPos;
					Blocks[realY * Width + realX].EnsureDecompressed();

					//Right leaf
					int rightX = realX < Width - 1 ? realX + 1 : 0;
					Blocks[realY * Width + rightX].EnsureDecompressed();

					if (realY + 1 < Height)
					{
						Blocks[(realY + 1) * Width + realX].EnsureDecompressed();
						Blocks[(realY + 1) * Width + rightX].EnsureDecompressed();
					}

					DrawLeaf(xPos * 32, yPos * 32, Blocks[realY * Width + realX], ret);

					Blocks[realY * Width + rightX].DiscardData();
					if (realY + 1 < Height)
					{
						Blocks[(realY + 1) * Width + realX].DiscardData();
						Blocks[(realY + 1) * Width + rightX].DiscardData();
					}
					Blocks[realY * Width + realX].DiscardData();
				}
			}

			return ret;
		}
		public void DrawTrueColoredGrayscaleLayer(Graphics g, Rectangle bounds)
		{
			DrawTrueColoredGrayscaleLayer(g, bounds, 0, 0);
		}
		public void DrawTrueColoredGrayscaleLayer(Graphics g, Rectangle bounds, int xStart, int yStart)
		{
			int xmax = bounds.Right / 32 + 1;
			int ymax = bounds.Bottom / 32 + 1;
			xmax += xStart;
			ymax += yStart;
			if (xmax > Width) xmax = Width;
			if (ymax > Height) ymax = Height;

			DrawTrueColoredGrayscaleLayer(g, bounds, xStart, yStart, xmax, ymax);
		}
		public void DrawTrueColoredGrayscaleLayer(Graphics g, Rectangle bounds, int xStart, int yStart, int xMax, int yMax)
		{
			int xmin = bounds.Left / 32 - 1; //Counted in blocks
			int xmax = bounds.Right / 32 + 1;
			int ymin = bounds.Top / 32 - 1;
			int ymax = bounds.Bottom / 32 + 1;
			xmin += xStart;
			xmax += xStart;
			ymin += yStart;
			ymax += yStart;
			if (xmin < 0) xmin = 0;
			if (xmax > Width) xmax = Width;
			if (ymin < 0) ymin = 0;
			if (ymax > Height) ymax = Height;
			if (xMax > xmax) xMax = xmax;
			if (yMax > ymax) yMax = ymax;

			ImageData<RgbData> temp = new ImageData<RgbData>(32 * (xmax - xmin), 32 * (ymax - ymin));
			for (int xPos = xmin; xPos < xMax; xPos++)
			{
				for (int yPos = ymin; yPos < yMax; yPos++)
				{
					this[xPos, yPos].EnsureDecompressed();

					//Right leaf
					int rightX = xPos < Width - 1 ? xPos + 1 : 0;
					this[rightX, yPos].EnsureDecompressed();

					if (yPos + 1 < Height)
					{
						this[xPos, yPos + 1].EnsureDecompressed();
						this[rightX, yPos + 1].EnsureDecompressed();
					}

					//ImageData<RgbData> temp = new ImageData<RgbData>(32, 32);
					DrawLeaf((xPos - xmin) * 32, (yPos - ymin) * 32, this[xPos, yPos], temp);
					//g.DrawImageUnscaled(temp.InternalBitmap, (xPos - xStart) * 32, (yPos - yStart) * 32);

					this[rightX, yPos].DiscardData();
					if (yPos + 1 < Height)
					{
						this[xPos, yPos + 1].DiscardData();
						this[rightX, yPos + 1].DiscardData();
					}
					this[xPos, yPos].DiscardData();
				}
			}
			g.DrawImageUnscaled(temp.InternalBitmap, (xmin - xStart) * 32, (ymin - yStart) * 32);
		}
		#endregion

		#endregion

		#region IEquatable<Lightmap> Members

		public bool Equals(Lightmap other)
		{
			if (this.Width != other.Width)
				return false;
			if (this.Height != other.Height)
				return false;
			for (int i = 0; i < Blocks.Length; i++)
				if (this.Blocks[i] != other.Blocks[i])
					return false;

			return true;
		}

		#endregion

		#region Writing
		public void Write()
		{
			using (BinaryStream outStream = new BinaryStream(path))
			{
				if (Settings.LmOffsetsAreFourBytes)
					Write4ByteOffsets(outStream);
				else
					Write3ByteOffsets(outStream);
				foreach (MapBlock block in Blocks)
				{
					outStream.Write(block);
				}
			}
		}

		private void Write4ByteOffsets(BinaryStream outStream)
		{
			outStream.WriteInt32(0);
			int currentPosition = 0;
			foreach (MapBlock block in Blocks)
			{
				currentPosition += block.StreamLength;
				outStream.WriteInt32(currentPosition);
			}
		}
		private void Write3ByteOffsets(BinaryStream outStream)
		{
			outStream.WriteBytes(0, 0, 0);
			int currentPosition = 0;
			foreach (MapBlock block in Blocks)
			{
				currentPosition += block.StreamLength;
				outStream.WriteByte((byte)(currentPosition & 255));
				outStream.WriteByte((byte)((currentPosition >> 8) & 255));
				outStream.WriteByte((byte)((currentPosition >> 16) & 255));
			}
			while (outStream.Position % 2 != 0)
				outStream.WriteByte();
		}
		#endregion

		#region Tiles
		private const int BlocksInRow = 256;
		private Point[,] tiles;

		private ImageData<GrayscaleData> tilesPic;
		private ColorPalette palette;

		public void LoadTileData()
		{
			if (tiles != null)
				return;

			tiles = new Point[Width, Height];

			if (!CoreLib.Io.File.Exists(Global.RootPath + @"\index.tbl"))
				return;
			if (!CoreLib.Io.File.Exists(Global.RootPath + @"\tiles.bmp"))
				return;

			Bitmap tilesImg = (Bitmap)Bitmap.FromFile(Global.RootPath + @"\tiles.bmp");
			this.palette = tilesImg.Palette;
			tilesPic = ImageData<GrayscaleData>.ToData(tilesImg);
			tilesImg.Dispose();
			tilesImg = null;
			palette.Entries[0] = Color.FromArgb(0, 255, 255, 255);
			tilesPic.InternalBitmap.Data.Palette = palette;

			//tilesImg.MakeTransparent(Color.FromArgb(0, 255, 0));

			//tilesPic = ImageData<ArgbData>.ToData(tilesImg);

			using (BinaryStream index = new BinaryStream(new FileStream(Global.RootPath + @"\index.tbl", FileMode.Open, FileAccess.Read)))
			{
				for (int x = 0; x < Width; x++)
				{
					for (int y = 0; y < Height; y++)
					{
						int val;
						if (Settings.IndexTblValuesAreFourBytes)
							val = index.ReadInt32();
						else
							val = index.ReadUInt16();
						if (val == 0)
							continue;
						int column = val % BlocksInRow;
						int row = val / BlocksInRow;
						tiles[x, y] = new Point(column, row);
					}
				}
			}
		}
		public ImageData<GrayscaleData> CreateTilesImage(int xBlock, int yBlock, int width, int height)
		{
			return CreateTilesImage(xBlock, yBlock, width, height, new ImageData<GrayscaleData>(width * 32, height * 32));
		}
		private ImageData<GrayscaleData> CreateTilesImage(int xBlock, int yBlock, int width, int height, 
			ImageData<GrayscaleData> canvas)
		{
			canvas.FillRectangle(0, 0, canvas.Width, canvas.Height, new GrayscaleData(0));
			canvas.InternalBitmap.Data.Palette = palette;

			if (tilesPic == null)
				return canvas;

			for (int x = 0; x < width; x++)
			{
				for (int y = 0; y < height; y++)
				{
					Point source = tiles[xBlock + x, yBlock + y];
					if (source.X == 0 && source.Y == 0)
						continue;
					ImageData<GrayscaleData>.Copy(tilesPic, source.X * 32, source.Y * 32, canvas, x * 32, y * 32, 32, 32);
				}
			}
			return canvas;
		}
		public void DrawTilesImage(Graphics g, Rectangle bounds)
		{
			DrawTilesImage(g, bounds, 0, 0);
		}
		public void DrawTilesImage(Graphics g, Rectangle bounds, int xStart, int yStart)
		{
			if (tiles == null)
				return;

			int xmin = bounds.Left / 32 - 1; //Counted in blocks
			int xmax = bounds.Right / 32 + 1;
			int ymin = bounds.Top / 32 - 1;
			int ymax = bounds.Bottom / 32 + 1;
			xmin += xStart;
			xmax += xStart;
			ymin += yStart;
			ymax += yStart;
			if (xmin < 0) xmin = 0;
			if (xmax > Width) xmax = Width;
			if (ymin < 0) ymin = 0;
			if (ymax > Height) ymax = Height;

			for (int y = ymin; y < ymax; y++)
			{
				for (int x = xmin; x < xmax; x++)
				{
					if (tiles[x, y].X == 0 && tiles[x, y].Y == 0)
						continue;
					ImageData<GrayscaleData> temp = new ImageData<GrayscaleData>(32, 32);
					temp.InternalBitmap.Data.Palette = palette;
					ImageData<GrayscaleData>.Copy(tilesPic, tiles[x, y].X * 32, tiles[x, y].Y * 32, temp, 0, 0, 32, 32);
					g.DrawImageUnscaled(temp.InternalBitmap, (x - xStart) * 32, (y - yStart) * 32, 32, 32);
				}
			}
		}
		#endregion
	}
	public enum LightmapOption
	{
		LightMap1,
		LightMap2,
		LightMap3,
		LightMap4
	}
}