﻿//#define WEAKREF
//#define TRYSHADOWBUFFER

using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using CoreLib;
using CoreLib.Graphics;
using CoreLib.Io;

namespace MapUtility
{
	public sealed class MapBlock : Leaf, IStreamable
	{
		#region Important Fields
		public ushort X { get; private set; }
		public ushort Y { get; private set; }
		public Lightmap WholeMap { get; set; }

		byte[] rawData;
		List<PixelOwnership> provinceIds;
		List<Leaf> drawnLeaves;

#if TRYSHADOWBUFFER
		//Will be 63 bytes
		//0-31 is right to left on top
		//31-62 is top to bottom on left
		private byte[] shadowBuffer;
#endif
#if WEAKREF
		private WeakReference provinceIdsPtr = new WeakReference(null);
		private WeakReference drawnLeavesPtr = new WeakReference(null);
		private WeakReference childrenPtr = new WeakReference(null); 
#endif
		#endregion

		#region Constructors
		public MapBlock(ushort x, ushort y, Lightmap wholeGrid, byte[] rawData)
			: this(x, y, wholeGrid)
		{
			this.rawData = rawData;
		}
		public MapBlock(ushort x, ushort y, Lightmap wholeGrid)
			: base(5, (ushort)(x * 32), (ushort)(y * 32))
		{
			this.WholeMap = wholeGrid;
			this.X = x;
			this.Y = y;
		}
		#endregion

		#region Decompression
#if !WEAKREF
		private bool decompressed;
#endif
		public void EnsureDecompressed()
		{
#if !WEAKREF
					if (decompressed)
				return;
  
#endif
#if WEAKREF
			if (provinceIds != null && Children != null && drawnLeaves != null)
				return;

			drawnLeaves = (List<Leaf>)drawnLeavesPtr.Target;
			provinceIds = (List<PixelOwnership>)provinceIdsPtr.Target;
			Children = (Leaf[])childrenPtr.Target;

			if (drawnLeaves != null && provinceIds != null && Children != null)
				return;
#endif

			int Offset = 0;
			drawnLeaves = new List<Leaf>();
#if !WEAKREF
			decompressed = true;
#endif
			if (Settings.IdsAreNewWay)
				GetProvinceIdsNewWay(ref Offset);
			else
				GetProvinceIdsOldWay(ref Offset);
			GetLeaves(ref Offset);
			GetLeafOwnership(ref Offset);
			GetColors(ref Offset);
			
#if WEAKREF
			provinceIdsPtr.Target = provinceIds;
			drawnLeavesPtr.Target = drawnLeaves;
			childrenPtr.Target = Children;
#endif
		}
		private void GetProvinceIdsNewWay(ref int Offset)
		{
			/*15	Ending bit - Same as standard IDs
			14	Special flag - On to show that this is a special ID
			13--------	|----------------------------------------------------------------------
			12	| Index1 - subtract four from this to get the index
			11	| of a value in the list of province IDs for the block
			10	| The value is an actual province, and when this
			9	| special province is clicked, that value is selected
			8----------	|-----------------------------------------------------------------------
			7----------	|-----------------------------------------------------------------------
			6	| Index2 - An index into the first province's list of 
			5	| adjacent provinces in adjacent.tbl
			4	|
			3	|
			2	|
			1----------	|-----------------------------------------------------------------------
			0	Inner/Outer border flag
			 */
			//SpecialProvs.Clear();
			provinceIds = new List<PixelOwnership>();
			while (true)
			{
				ushort u = (ushort)(rawData[Offset] | (rawData[Offset + 1] << 8));
				ushort tempValue = (ushort)(u & 0x7fff);
				if (tempValue < 3000)
					provinceIds.Add(new PixelOwnership(tempValue));
				else
				{
					bool IsTypeOneBorder = (tempValue & 1) != 0;
					int index1 = (tempValue >> 8) & 63;
					int index2 = (tempValue >> 1) & 63;
					ushort owner = provinceIds[index1 - 4].OwnerId;
					provinceIds.Add(new PixelOwnership(owner, IsTypeOneBorder, Global.AdjacentTable[owner, index2]));
				}
				Offset += 2;
				if ((u >> 15) != 0) break;
			}
		}
		private void GetProvinceIdsOldWay(ref int Offset)
		{
			/*15	Ending bit - Same as standard IDs
			14--------|----------------------------------------------------------------------
			13	| Index1 - subtract four from this to get the index
			12	| of a value in the list of province IDs for the block
			11	| The value is an actual province, and when this
			10	| special province is clicked, that value is selected
			9-----------|---------------------------------------------------------------------
			8----------	|---------------------------------------------------------------------
			7	| River flag - River if not bits set to 15
			6	| 
			5-----------|-----------------------------------------------------------------------
			4-----------|-----------------------------------------------------------------------
			3	| Index2 - An index into the first province's list of 
			2	| adjacent provinces in adjacent.tbl
			1----------	|-----------------------------------------------------------------------
			0	Inner/Outer border flag
			 */
			provinceIds = new List<PixelOwnership>();

			while (true)
			{
				ushort u = (ushort)(rawData[Offset] | (rawData[Offset + 1] << 8));
				ushort tempValue = (ushort)(u & 0x7fff);
				if (tempValue < 1100)
					provinceIds.Add(new PixelOwnership(tempValue));
				else
				{
					bool IsTypeOneBorder = (tempValue & 1) != 0;
					int index1 = (tempValue >> 9) & 63;
					int index2 = (tempValue >> 1) & 15;
					ushort owner = provinceIds[index1 - 4].OwnerId;
					provinceIds.Add(new PixelOwnership(owner, IsTypeOneBorder, Global.AdjacentTable[owner, index2]));
				}
				Offset += 2;
				if ((u >> 15) != 0) break;
			}
		}
		private void GetLeaves(ref int Offset)
		{
			int bitIndex = 0;
			byte[] source = new byte[rawData.Length - Offset];
			Array.Copy(rawData, Offset, source, 0, rawData.Length - Offset);
			BitArray bits = new BitArray(source);
			DoLeaf(this, ref bitIndex, bits,ref Offset);
		}
		private void DoLeaf(Leaf leaf, ref int bitIndex, BitArray bits, ref int Offset)
		{
			if (bitIndex % 8 == 0)
				Offset++;

			if (bits[bitIndex++])
			{
				if (leaf.Level == 1)
				{
					for (int i = 0; i < 4; i++)
					{
						Leaf child = new Leaf(leaf, i);
						leaf.AddChild(child);
						drawnLeaves.Add(child);
					}
				}
				else
				{
					for (int i = 0; i < 4; i++)
					{
						Leaf child = new Leaf(leaf, i);
						leaf.AddChild(child);
						DoLeaf(child, ref bitIndex, bits, ref Offset);
					}
				}
			}
			else
				drawnLeaves.Add(leaf);
		}
		private void GetColors(ref int Offset)
		{
			int count = this.HasChildren ? this.drawnLeaves.Count : 1;

			GrayscaleData Color = new GrayscaleData();
			for (int i = 0; i < count; i++)
			{
				switch (i % 4)
				{
					case 0:
						Color = (rawData[Offset] & 63);
						break;
					case 1:
						Color = (((rawData[Offset] >> 6) | rawData[Offset + 1] << 2) & 63);
						break;
					case 2:
						Color = (((rawData[Offset + 1] >> 4) | rawData[Offset + 2] << 4) & 63);
						break;
					case 3:
						Color = (rawData[Offset + 2] >> 2 & 63);
						Offset += 3;
						break;
				}
				this.drawnLeaves[i].GrayscaleColor = Color;
			}
		}
		private void GetLeafOwnership(ref int Offset)
		{
			if (provinceIds.Count == 1)
				this.SetAllOwners(provinceIds[0]);
			else if (provinceIds.Count == 2)
				GetBitOwnership(1, ref Offset);
			else if (provinceIds.Count <= 4)
				GetBitOwnership(2, ref Offset);
			else if (provinceIds.Count <= 16)
				GetBitOwnership(4, ref Offset);
			else //17 - 256
				GetBitOwnership(8, ref Offset);
		}
		private void GetBitOwnership(int width, ref int Offset)
		{
			int pos = 0;
			for (int iterator = 0; iterator < drawnLeaves.Count; iterator++)
			{
				this.drawnLeaves[iterator].OwnerId = provinceIds[(rawData[Offset] >> pos) & (((1 << width) - 1))];
				pos += width;
				if (pos == 8)
				{
					pos = 0;
					Offset++;
				}
			}
			if (pos != 0)
				Offset++;
		}
		public void DiscardData()
		{
#if !WEAKREF
			decompressed = false;
#endif
			Children = null;
			provinceIds = null;
			drawnLeaves = null;
		}
		#endregion

		public override string ToString()
		{
			return string.Format("X={0}, Y={1}, {2}", X, Y, base.ToString());
		}

		#region Province ids public interface
		/// <summary>
		/// Confirms that all province ids in the block satsify a certain condition.
		/// </summary>
		/// <param name="test">The test whose condition must be met.</param>
		/// <returns>True if all conditions are met.</returns>
		public bool ForAllProvinceIds(Predicate<PixelOwnership> test)
		{
			return provinceIds.All(pixel => test(pixel));
		}
		public void SetProvinceIds(IEnumerable<PixelOwnership> source)
		{
			if (provinceIds == null)
				provinceIds = new List<PixelOwnership>();
			provinceIds.Clear();
			provinceIds.Capacity = source.Count();
			foreach (PixelOwnership po in source)
				provinceIds.Add(po);
			provinceIds.Sort();
		}
		#endregion

		#region Compression
		public void FinishedToRaw()
		{
			List<byte> newData = new List<byte>(50);
			if (Settings.IdsAreNewWay)
				WriteProvinceIdsNewWay(newData);
			else
				WriteProvinceIdsOldWay(newData);
			WriteQuadTree(newData);
			WriteOwnership(newData);
			WriteGrayscale(newData);

			rawData = newData.ToArray();
		}

		private void WriteProvinceIdsNewWay(List<byte> newData)
		{
			foreach (PixelOwnership ownership in provinceIds)
			{
				ushort u = 
					ownership.ToUshortNewWay(
					provinceIds.FindIndex(pixel => pixel.OwnerId == ownership.OwnerId && !pixel.IsBorder), false);
				newData.Add((byte)(u & 255));
				newData.Add((byte)((u >> 8) & 255));
			}
			newData[newData.Count - 1] |= (1 << 7);
		}
		private void WriteProvinceIdsOldWay(List<byte> newData)
		{
			foreach (PixelOwnership ownership in provinceIds)
			{
				ushort u =
					ownership.ToUshortOldWay(
					provinceIds.FindIndex(pixel => pixel.OwnerId == ownership.OwnerId && !pixel.IsBorder), false);
				newData.Add((byte)(u & 255));
				newData.Add((byte)((u >> 8) & 255));
			}
			newData[newData.Count - 1] |= (1 << 7);
		}
		private void WriteQuadTree(List<byte> newData)
		{
			List<bool> prototree = new List<bool>(10);
			AppendNumbers(prototree);

			BitArray quadtree = new BitArray(prototree.ToArray());

			int numNewBytes = (quadtree.Count + 7) / 8;
			for (int i = 0; i < numNewBytes; i++)
			{
				byte b = 0;
				int last = i == numNewBytes - 1 ?
					(quadtree.Count % 8 == 0 ? 8 : quadtree.Count % 8)
					: 8;
				for (int pos = 0; pos < last; pos++)
				{
					b |= (byte)((quadtree[i * 8 + pos] ? 1 : 0) << pos);
				}
				newData.Add(b);
			}
		}
		private void WriteOwnership(List<byte> newData)
		{
			if (provinceIds.Count != 1)
			{
				List<int> ownerIdPositions = new List<int>(50);
				AppendOwnerIdInfo(provinceIds, ownerIdPositions);

				if (provinceIds.Count == 2)
				{
					WriteOwnerIdInfo(newData, ownerIdPositions, 8);
				}
				else if (provinceIds.Count <= 4)
				{
					WriteOwnerIdInfo(newData, ownerIdPositions, 4);
				}
				else if (provinceIds.Count <= 16)
				{
					WriteOwnerIdInfo(newData, ownerIdPositions, 2);
				}
				else //if (ProvinceIds.Count <= 256)
				{
					foreach (int val in ownerIdPositions)
						newData.Add((byte)val);
				}
			}
		}
		private void WriteGrayscale(List<byte> newData)
		{
			List<byte> Colors = new List<byte>(50);
			AppendColorInfo(Colors);

			while (Colors.Count % 4 != 0)
				Colors.Add(0);

			for (int i = 0; i < Colors.Count / 4; i++)
			{
				byte first;// = 0;
				byte second;// = 0;
				byte third;// = 0;

				first = Colors[i * 4];
				first |= (byte)((Colors[i * 4 + 1] & 3) << 6);
				second = (byte)(Colors[i * 4 + 1] >> 2);
				second |= (byte)((Colors[i * 4 + 2] & 15) << 4);
				third = (byte)(Colors[i * 4 + 2] >> 4);
				third |= (byte)(Colors[i * 4 + 3] << 2);

				newData.Add(first);
				newData.Add(second);
				newData.Add(third);
			}
		}
		private static void WriteOwnerIdInfo(List<byte> newData, List<int> ownerIdPositions, int numInByte)
		{
			int length = (ownerIdPositions.Count + numInByte - 1) / numInByte;
			int ownershipIndexWidth = 8 / numInByte;
			int AndVal = (1 << ownershipIndexWidth) - 1;
			for (int i = 0; i < length; i++)
			{
				byte b = 0;
				int last = i == length - 1 ?
					(ownerIdPositions.Count % numInByte == 0 ? numInByte : ownerIdPositions.Count % numInByte)
					: numInByte;
				for (int pos = 0; pos < last; pos++)
				{
					b |= (byte)((ownerIdPositions[i * numInByte + pos] & AndVal) << (pos * ownershipIndexWidth));
				}
				newData.Add(b);
			}
		}

		public void Write(IStreamWriter stream)
		{
			stream.WriteBytes(rawData);
		}
		public int StreamLength { get { return rawData.Length; } }
		#endregion

		#region Shadow buffer
#if TRYSHADOWBUFFER
		public byte GetTopValue(int position)
		{
			if (shadowBuffer == null)
				LoadShadowBuffer();
			return shadowBuffer[31 - position];
		}
		public byte GetLeftValue(int position)
		{
			if (shadowBuffer == null)
				LoadShadowBuffer();
			return shadowBuffer[31 + position];
		}
		public byte TopLeftValue
		{
			get
			{
				if (shadowBuffer == null)
					LoadShadowBuffer();
				return shadowBuffer[31];
			}
		}
		private void LoadShadowBuffer()
		{
			this.Decompress();
			shadowBuffer = new byte[63];

			//Right leaf
			int rightX = X < wholeGrid.Width - 1 ? X + 1 : 0;
			wholeGrid[Y, rightX].Decompress();

			if (Y + 1 < wholeGrid.Height)
			{
				wholeGrid[Y + 1, X].Decompress();
				wholeGrid[Y + 1, rightX].Decompress();
			}

			byte[,] buffer = new byte[32, 32];
			DrawLeaf(0, 0, this, buffer);

			//Top values
			for (int i = 0; i < 32; i++)
			{
				shadowBuffer[31 - i] = buffer[i, 0];
			}

			//Left values
			for (int i = 0; i < 32; i++)
			{
				shadowBuffer[i + 31] = buffer[0, i];
			}

			wholeGrid[Y, rightX].DiscardData();
			if (Y + 1 < wholeGrid.Height)
			{
				wholeGrid[Y + 1, X].DiscardData();
				wholeGrid[Y + 1, rightX].DiscardData();
			}
			this.DiscardData();
		}
		private static void DrawLeaf(int x, int y, Leaf l, byte[,] buffer)
		{
			#region If has children
			if (l.HasChildren)
			{
				int shift = l.Shift >> 1;
				DrawLeaf(x, y, l.TopLeftChild, buffer);
				DrawLeaf(x + shift, y, l.TopRightChild, buffer);
				DrawLeaf(x, y + shift, l.BottomLeftChild, buffer);
				DrawLeaf(x + shift, y + shift, l.BottomRightChild, buffer);
				return;
			}
			#endregion
			#region If is single pixel
			if (l.Level == 0)
			{
				buffer[x, y] = l.DeepGrayscaleColor;
				return;
			}
			#endregion

			#region Declare neighbor leaves
			Leaf bottomRightLeaf = l.BottomRightNeighbor;
			Leaf bottomLeaf = l.BottomNeighbor;
			Leaf rightLeaf = l.RightNeighbor;
			#endregion

			#region If some neighbors are missing
			if (bottomRightLeaf == null || bottomLeaf == null || rightLeaf == null)
			{
				for (int x1 = 0; x1 < l.Shift; x1++)
				{
					for (int y1 = 0; y1 < l.Shift; y1++)
					{
						buffer[x + x1, y + y1] = l.DeepGrayscaleColor;
					}
				}
				return;
			}
			#endregion

			#region Declare grayscale values
			byte topLeft = l.DeepGrayscaleColor;
			byte right = rightLeaf.DeepGrayscaleColor;
			byte bottom = bottomLeaf.DeepGrayscaleColor;
			byte bottomRight = bottomRightLeaf.DeepGrayscaleColor;
			#endregion

			#region If all values are same
			if (topLeft == right && right == bottom && bottom == bottomRight) //Since all are same, simple filling in
			{
				for (int x1 = 0; x1 < l.Shift; x1++)
				{
					for (int y1 = 0; y1 < l.Shift; y1++)
					{
						buffer[x + x1, y + y1] = l.DeepGrayscaleColor;
					}
				}
				return;
			}
			#endregion

			byte[,] data = new byte[2, l.Shift + 1];
			data[0, 0] = topLeft;
			data[0, l.Shift] = bottom;
			data[1, l.Shift] = bottomRight;
			data[1, 0] = right;

			//int realMaxDiff = MaxDiff;
			//Is int because shifts can be negative, and sbytes might not have enough range
			int firstShift = bottom - topLeft;
			int secondShift = bottomRight - right;
			for (int i2 = 1; i2 < l.Shift; i2++)
			{
				data[0, i2] = (byte)(topLeft + (firstShift * i2 + 0.5) / l.Shift);
				data[1, i2] = (byte)(right + (secondShift * i2 + 0.5) / l.Shift);
			}

			for (int row = 0; row < l.Shift; row++)
			{
				int valShift = data[1, row] - data[0, row];
				for (int i2 = 0; i2 < l.Shift; i2++)
				{
					buffer[x + i2, y + row] = (byte)(data[0, row] + (valShift * i2 + 0.5) / l.Shift);
				}
			}

		}
#endif
		#endregion
	}
}