﻿using System;
using System.Diagnostics;
using System.Globalization;
using CoreLib;
using CoreLib.Io;
using System.Runtime.InteropServices;

namespace MapUtility
{
	/// <summary>
	/// Has the information for ownership of one pixel in a block.
	/// </summary>
	[StructLayout(LayoutKind.Explicit, Size=5)]
	public struct PixelOwnership : IEquatable<PixelOwnership>, IComparable<PixelOwnership>, IStreamable
	{
		//The data structure will be as follows
		//Two bytes for the owner ID
		//One byte for the flags
		//If flag 0 is on, another two bytes for the other owner

		#region Constructors
		public PixelOwnership(ushort ownerId)
			: this()
		{
			this.OwnerId = ownerId;
		}
		public PixelOwnership(ushort ownerId, bool isTypeOne, ushort otherOwner)
			: this()
		{
			this.OwnerId = ownerId;
			this.IsTypeOne = isTypeOne;
			this.OtherOwner = otherOwner;
			this.IsBorder = true;
		}
		#endregion

		#region Data
		//WARNING: If make these properties, must consider data layout
		[FieldOffset(0)]
		public ushort OwnerId;
		[FieldOffset(2)]
		private ByteFlag data;
		public bool IsBorder
		{
			[DebuggerHidden]
			get { return data[0]; }
			[DebuggerHidden]
			set { data[0] = value; }
		}
		public bool IsTypeOne //Type 1 is blue (And is on)
		{
			[DebuggerHidden]
			get { return data[1]; }
			[DebuggerHidden]
			set { data[1] = value; }
		}
		/// <summary>
		/// If is a border tile, this is what the other province ID is
		/// </summary>
		[FieldOffset(3)]
		public ushort OtherOwner;
		#endregion

		#region Comparison and equality
		[DebuggerHidden]
		public override bool Equals(object obj)
		{
			try
			{
				return this.Equals((PixelOwnership)obj);
			}
			catch (InvalidCastException)
			{
				return false;
			}
		}
		[DebuggerHidden]
		public bool Equals(PixelOwnership other)
		{
			if (this.IsBorder && other.IsBorder)
				return this.OtherOwner == other.OtherOwner && this.IsTypeOne == other.IsTypeOne && this.OwnerId == other.OwnerId;
			if (!this.IsBorder && !other.IsBorder)
				return this.OwnerId == other.OwnerId;
			return false;
		}
		[DebuggerHidden]
		public static bool operator ==(PixelOwnership first, PixelOwnership second)
		{
			return first.Equals(second);
		}
		[DebuggerHidden]
		public static bool operator !=(PixelOwnership first, PixelOwnership second)
		{
			return !first.Equals(second);
		}
		[DebuggerHidden]
		public override int GetHashCode()
		{
			return OwnerId.GetHashCode() ^ data.GetHashCode() ^ OtherOwner.GetHashCode();
		}
		public int CompareTo(PixelOwnership other)
		{
			if (this.OwnerId != other.OwnerId)
				return this.OwnerId.CompareTo(other.OwnerId);

			//Ids are equal, next is who is not a border
			if (!this.IsBorder && other.IsBorder)
				return -1;
			if (this.IsBorder && !other.IsBorder)
				return 1;

			//They are both borders, lets look at the otherOwner ids
			if (this.OtherOwner != other.OtherOwner)
				return this.OtherOwner.CompareTo(other.OtherOwner);

			//Same thing, must look at border type
			if (this.IsTypeOne && !other.IsTypeOne)
				return 1;
			if (!this.IsTypeOne && other.IsTypeOne)
				return -1;

			return 0; //They are the complete same
		}
		public static bool operator >(PixelOwnership first, PixelOwnership second)
		{
			return first.CompareTo(second) > 0;
		}
		public static bool operator <(PixelOwnership first, PixelOwnership second)
		{
			return first.CompareTo(second) < 0;
		}
		public static bool operator >=(PixelOwnership first, PixelOwnership second)
		{
			return first.CompareTo(second) >= 0;
		}
		public static bool operator <=(PixelOwnership first, PixelOwnership second)
		{
			return first.CompareTo(second) <= 0;
		}
		#endregion

		/// <summary>
		/// 
		/// </summary>
		/// <param name="ownerPos">If this is a border, the ID of the owner province. If is not a border, we don't care.</param>
		/// <param name="IsLast"></param>
		/// <returns></returns>
		public ushort ToUshortNewWay(int ownerPos, bool isLast)
		{
			if (!IsBorder)
			{
				if (isLast)
				{
					return (ushort)(OwnerId | (1 << 15));
				}

				return OwnerId;
			}

			//It IS a border
			int index1 = ownerPos + 4;
			int index2 = Global.AdjacentTable[OwnerId].IndexOf(OtherOwner);
			if (index2 == -1) //The id was not found, we must pretend it is not a border
			{
				this.IsBorder = false;
				return ToUshortNewWay(0, isLast);
			}

			int result = (1 << 14);
			if (IsTypeOne)
			{
				result |= 1;
			}

			result |= (index1 << 8);
			result |= (index2 << 1);
			if (isLast)
			{
				result |= (1 << 15);
			}

			return (ushort)result;
		}
		/// <summary>
		/// 
		/// </summary>
		/// <param name="ownerPos">If this is a border, the ID of the owner province. If is not a border, we don't care.</param>
		/// <param name="IsLast"></param>
		/// <returns></returns>
		public ushort ToUshortOldWay(int ownerPos, bool isLast)
		{
			if (!IsBorder)
			{
				if (isLast)
				{
					return (ushort)(this.OwnerId | (1 << 15));
				}

				return this.OwnerId;
			}

			//It IS a border
			int index1 = ownerPos + 4;
			int index2 = Global.AdjacentTable[OwnerId].IndexOf(OtherOwner);

			if (index2 == -1) //ID not found, pretend it is not a border
			{
				this.IsBorder = false;
				return this.ToUshortOldWay(0, isLast);
			}

			int result = 0;//(1 << 14);
			if (IsTypeOne)
				result |= 1;
			result |= (index1 << 9);
			result |= (index2 << 1);
			result |= (15 << 5);//River flag
			if (isLast)
				result |= (1 << 15);
			return (ushort)result;
		}

		/*public RgbData ToColor()
		{
			return Global.IdToColor[this.ownerId];
		}*/

		#region Streaming
		public const int Length = 5;

		[DebuggerHidden]
		public void Write(IStreamWriter stream)
		{
			stream.WriteUInt16(OwnerId);
			stream.WriteByte(data);
			stream.WriteUInt16(OtherOwner);
		}
		[DebuggerHidden]
		public unsafe static PixelOwnership Read(IStreamReader stream)
		{
			fixed (byte* p = &stream.ReadBytes(5)[0])
			{
				return *(PixelOwnership*)p;
			}
			//PixelOwnership ret = new PixelOwnership();
			//ret.OwnerId = stream.ReadUInt16();
			//ret.data = stream.ReadByte();
			//ret.OtherOwner = stream.ReadUInt16();
			//return ret;
		}
		#endregion

		public override string ToString()
		{
			if (!this.IsBorder)
				return "Id: " + OwnerId.ToString(NumberFormatInfo.CurrentInfo);
			return "Id: " + OwnerId.ToString(NumberFormatInfo.CurrentInfo) + 
				" Type: " + (IsTypeOne ? 1 : 0) + " other: " + 
				OtherOwner.ToString(NumberFormatInfo.CurrentInfo);
		}
	}
}