﻿using System;
using System.Collections.Generic;
using CoreLib;
using CoreLib.Io;
using StreamWriter = System.IO.StreamWriter;

namespace MapUtility
{
	public static class Settings
	{
		#region Basic
		private static Dictionary<string, string> data = new Dictionary<string, string>();

		static Settings()
		{
			if (File.Exists("settings.txt"))
			{
				List<string> vals = File.ReadAndCleanLines("settings.txt");
				foreach (string s in vals)
				{
					string[] parts = s.Split('=');
					if (parts.Length != 2)
						continue;
					parts[0].ChangeTrim();
					parts[1].ChangeTrim();

					data.Add(parts[0], parts[1]);
				}

				//Load map property values
				if (!int.TryParse(GetValue(Lightmap1WidthKey), out Lm1Width))
				{
					Lm1Width = 936;
					SetValue(Lightmap1WidthKey, Lm1Width.ToString());
				}

				if (!int.TryParse(GetValue(Lightmap1HeightKey), out Lm1Height))
				{
					Lm1Height = 360;
					SetValue(Lightmap1HeightKey, Lm1Height.ToString());
				}

				int numLightmaps;
				if (!int.TryParse(GetValue(NumLightmapsKey), out numLightmaps))
				{
					HasFourLightmaps = true;
					SetValue(NumLightmapsKey, 4.ToString());
				}
				else
				{
					HasFourLightmaps = numLightmaps == 4;
				}

				int offsetWidth;
				if (!int.TryParse(GetValue(LightmapOffsetWidthKey), out offsetWidth))
				{
					LmOffsetsAreFourBytes = true;
					SetValue(LightmapOffsetWidthKey, 4.ToString());
				}
				else
				{
					LmOffsetsAreFourBytes = offsetWidth == 4;
				}

				int indexTblWidth;
				if (!int.TryParse(GetValue(IndexTblValueWidthKey), out indexTblWidth))
				{
					IndexTblValuesAreFourBytes = true;
					SetValue(IndexTblValueWidthKey, 4.ToString());
				}
				else
					IndexTblValuesAreFourBytes = indexTblWidth == 4;

				string idsAreNewWayVal = GetValue(IdsAreNewWayKey);
				if (idsAreNewWayVal == "false")
					IdsAreNewWay = false;
				else
					IdsAreNewWay = true;
				
			}
			else
			{
				//Default values - will not work for CK
				Lm1Width = 936;
				Lm1Height = 360;
				HasFourLightmaps = true;
				LmOffsetsAreFourBytes = true;
				SetValue(Lightmap1WidthKey, Lm1Width.ToString());
				SetValue(Lightmap1HeightKey, Lm1Height.ToString());
				SetValue(NumLightmapsKey, 4.ToString());
				SetValue(LightmapOffsetWidthKey, 4.ToString());
				SetValue(IdsAreNewWayKey, "true");
				SetValue(IndexTblValueWidthKey, 4.ToString());
			}

			Lm2Width = Lm1Width / 2;
			Lm3Width = Lm2Width / 2;
			Lm4Width = Lm3Width / 2;

			Lm2Height = Lm1Height / 2;
			Lm3Height = Lm2Height / 2;
			Lm4Height = Lm3Height / 2;

			SmallestLmHeight = HasFourLightmaps ? Lm4Height : Lm3Height;
			SmallestLmWidth = HasFourLightmaps ? Lm4Width : Lm3Width;
		}

		public static string GetValue(string key)
		{
			string ret;
			if (!data.TryGetValue(key, out ret))
				return string.Empty;
			return ret;
		}
		public static void SetValue(string key, string value)
		{
			data[key] = value;
			Save();
		}
		public static bool TryGetValue(string key, out string value)
		{
			if (!data.TryGetValue(key, out value))
				return false;
			return true;
		}

		private static void Save()
		{
			using (StreamWriter sw = new StreamWriter("settings.txt", false))
			{
				foreach (KeyValuePair<string, string> item in data)
				{
					sw.Write(item.Key);
					sw.Write('=');
					sw.Write(item.Value);
					sw.WriteLine();
				}
			}
		}
		#endregion

		#region Max Province Number
		private const string MaxProvinceNumberKey = "MaxProvinceNumber";
		private static int? maxProvNumber = null;
		public static int QueryMaxProvinceNumber()
		{
			//HoI: 2608
			//Vicky: 2864
			//CK: 1003

			if (maxProvNumber != null)
				return maxProvNumber.Value;

			int value;
			if (!int.TryParse(GetValue(MaxProvinceNumberKey), out value) || value <= 0)
			{
				while (true)
				{
					Console.WriteLine("What is the max province number?");
					Console.WriteLine("(HoIDD 1.0-1.3a, HoIDDA 1.0-1.2: 2608)");
					Console.WriteLine("(Victoria 1.0-?, VR 1.0-1.1: 2864)");
					Console.WriteLine("(CK 1.0-?, CKDV 1.0-?: 1004");
					if (!int.TryParse(Console.ReadLine(), out value))
					{
						Console.WriteLine("Invalid number, try again!");
					}
					else
					{
						maxProvNumber = value;
						SetValue(MaxProvinceNumberKey, value.ToString());
						return maxProvNumber.Value;
					}
				}
			}
			else
			{
				maxProvNumber = value;
				return value;
			}

		}
		public static int? GetMaxProvNumber()
		{
			if (maxProvNumber != null)
				return maxProvNumber.Value;

			int value;
			if (int.TryParse(GetValue(MaxProvinceNumberKey), out value))
			{
				maxProvNumber = value;
				return value;
			}

			return null;
		}
		#endregion

		#region File Paths
		private const string GamePathKey = "GamePath";
		public static string GamePath
		{
			get
			{
				string val;
				if (!TryGetValue(GamePathKey, out val))
				{
					SetValue(GamePathKey, string.Empty);
					return string.Empty;
				}
				return val;
			}
			set
			{
				SetValue(GamePathKey, value);
			}
		}
		#endregion

		#region Map Info
		private const string Lightmap1WidthKey = "Lightmap1Width";
		private const string Lightmap1HeightKey = "Lightmap1Height";
		private const string NumLightmapsKey = "NumLightmaps";
		private const string LightmapOffsetWidthKey = "LightmapOffsetWidth";
		private const string IdsAreNewWayKey = "IdsAreNewWay";
		private const string IndexTblValueWidthKey = "IndexTblValueWidth";

		public static readonly bool HasFourLightmaps;
		public static readonly bool LmOffsetsAreFourBytes;
		public static readonly bool IdsAreNewWay;
		public static readonly bool IndexTblValuesAreFourBytes;

		public static readonly int Lm1Width;
		public static readonly int Lm1Height;
		public static readonly int Lm2Width;
		public static readonly int Lm2Height;
		public static readonly int Lm3Width;
		public static readonly int Lm3Height;
		public static readonly int Lm4Width;
		public static readonly int Lm4Height;

		public static readonly int SmallestLmWidth;
		public static readonly int SmallestLmHeight;
		#endregion
	}
}